"""Functionality for working multi-threaded code."""

import inspect
import logging
import threading
import time
from typing import Dict, Tuple

from .._internal.trace import trace_str


class DeadlockMonitor:
    """A monitor for deadlocked LoggingLocks."""

    timeout_sec: float
    sleep_sec: float
    locks: Dict[int, Tuple[float, int, str, str]]
    _lock: threading.Lock
    _thread: threading.Thread

    def __init__(self, timeout_sec: float, sleep_sec: float):
        self.timeout_sec = timeout_sec
        self.sleep_sec = sleep_sec
        self.locks = {}
        self._lock = threading.Lock()

        self._thread = threading.Thread(name="DeadlockMonitor", target=self._run, daemon=True)
        self._thread.start()
        setattr(self, "deadlock_monitor_thread", self._thread)

    def _run(self):
        while True:
            self._check_for_deadlocks()
            time.sleep(self.sleep_sec)

    def _check_for_deadlocks(self):
        with self._lock:
            t = time.time()

            for k, v in self.locks.items():
                if t - v[0] > self.timeout_sec:
                    self._log_deadlock()
                    return

    def _log_deadlock(self):
        t = time.time()
        msg = "A likely deadlock was detected!  Please create an issue at https://github.com/deephaven-examples/deephaven-ib/issues containing this error message\nOpen locks:\n"

        for k, v in self.locks.items():
            msg += f"age_sec={t-v[0]} lock_id={v[1]} name={v[2]}\n"

        msg += "\n\nStacks:\n\n"

        for k, v in self.locks.items():
            msg += f"age_sec={t-v[0]} lock_id={v[1]} name={v[2]}\n{v[3]}\n"

        logging.error(msg)

    def acquire(self, lock_id: int, name: str, stack: str) -> None:
        with self._lock:
            self.locks[lock_id] = (time.time(), lock_id, name, stack)

    def release(self, lock_id: int):
        with self._lock:
            # pop is used here instead of del, because there are instances where the locks are released multiple times
            self.locks.pop(lock_id, None)


_lock_id: int = 0
_lock: threading.Lock = threading.Lock()
_deadlock_monitor: DeadlockMonitor = DeadlockMonitor(3 * 60.0, 10.0)


def _next_lock_id() -> int:
    global _lock_id

    with _lock:
        _lock_id += 1
        return _lock_id


class LoggingLock(object):
    """A threading lock that logs lock acquisition and release."""

    name: str
    log_stack: bool

    def __init__(self, name: str, lock=None, log_level=logging.DEBUG, log_stack: bool = False):
        if lock is None:
            lock = threading.Lock()

        self.name = str(name)
        self.log_level = log_level
        self.lock = lock
        self.log_stack = log_stack
        self.id = _next_lock_id()
        self._log(f"{inspect.stack()[1][3]} created {self.name}")

    def _log(self, msg: str) -> None:
        if self.log_stack:
            msg = f"{msg}: lock_id={self.id} thread_id={threading.get_ident()}\n{trace_str()}"
        else:
            msg = f"{msg}: lock_id={self.id} thread_id={threading.get_ident()}"

        logging.log(self.log_level, msg)

    def acquire(self, blocking=True):
        self._log(f"{inspect.stack()[1][3]} trying to acquire {self.name}")

        if _deadlock_monitor:
            _deadlock_monitor.acquire(self.id, self.name, trace_str())

        ret = self.lock.acquire(blocking)

        if ret:
            self._log(f"{inspect.stack()[1][3]} acquired {self.name}")
        else:
            self._log(f"{inspect.stack()[1][3]} non-blocking acquire of {self.name} lock failed")

        return ret

    def release(self):
        self._log(f"{inspect.stack()[1][3]} releasing {self.name}")

        if _deadlock_monitor:
            _deadlock_monitor.release(self.id)

        self.lock.release()

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()
        return False  # True causes exceptions to be swallowed.  False causes exceptions to be handled.



"""An IB TWS client that produces Deephaven tables."""

import html
import json
import logging
import time
import types
# noinspection PyPep8Naming
import xml.etree.ElementTree as ET
from functools import wraps
from threading import Thread
from typing import Set, Optional

import decimal
from decimal import Decimal

from deephaven.table import Table
from deephaven import dtypes

from ibapi import news
from ibapi.client import EClient
from ibapi.commission_report import CommissionReport
from ibapi.common import *
from ibapi.contract import Contract, ContractDetails
from ibapi.execution import Execution, ExecutionFilter
from ibapi.order import Order
from ibapi.order_state import OrderState
from ibapi.ticktype import TickType, TickTypeEnum
from ibapi.wrapper import EWrapper
from ratelimit import limits, sleep_and_retry

from .contract_registry import ContractRegistry
from .ib_type_logger import *
from .order_id_queue import OrderIdEventQueue, OrderIdStrategy
from .requests import RequestIdManager
from .._internal.error_codes import load_error_codes
from .._internal.short_rates import load_short_rates
from .._internal.tablewriter import TableWriter
from ..time import unix_sec_to_j_instant

_error_code_message_map, _error_code_note_map = load_error_codes()
_news_msgtype_map: Dict[int, str] = {news.NEWS_MSG: "NEWS", news.EXCHANGE_AVAIL_MSG: "EXCHANGE_AVAILABLE",
                                     news.EXCHANGE_UNAVAIL_MSG: "EXCHANGE_UNAVAILABLE"}


# Rate limit is 50 per second.  Limiting to 45 per second.
@sleep_and_retry
@limits(calls=45, period=1)
def _check_rate_limit():
    """Empty function to limit the rate of calls to API."""
    pass


def _rate_limit_wrapper(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        _check_rate_limit()
        return func(*args, **kwargs)

    return wrapped


# noinspection PyPep8Naming
class IbTwsClient(EWrapper, EClient):
    """A client for communicating with IB TWS.

    Almost all of the methods in this class are listeners for EWrapper and should not be called directly by users of the class.
    """

    _table_writers: Dict[str, TableWriter]
    tables: Dict[str, Table]
    _thread: Thread
    contract_registry: ContractRegistry
    request_id_manager: RequestIdManager
    order_id_queue: OrderIdEventQueue
    _registered_market_rules: Set[str]
    _realtime_bar_sizes: Dict[TickerId, int]
    news_providers: List[str]
    _accounts_managed: Set[str]
    _order_id_strategy: OrderIdStrategy
    _read_only: bool
    _is_fa: bool

    def __init__(self, download_short_rates: bool, order_id_strategy: OrderIdStrategy, read_only: bool, is_fa: bool):
        EWrapper.__init__(self)
        EClient.__init__(self, wrapper=self)
        self._table_writers = IbTwsClient._build_table_writers()
        self._thread = None
        self.contract_registry = None
        self.request_id_manager = RequestIdManager()
        self.order_id_queue = None
        self._registered_market_rules = None
        self._realtime_bar_sizes = None
        self.news_providers = None
        self._accounts_managed = None
        self._order_id_strategy = order_id_strategy
        self._read_only = read_only
        self._is_fa = is_fa

        tables = {name: tw.table() for (name, tw) in self._table_writers.items()}

        if download_short_rates:
            tables["short_rates"] = load_short_rates()

        self.tables = dict(sorted(tables.items()))

    # wrap all method names starting with "req" with a rate limiter.
    def __getattribute__(self, name):
        attr = EClient.__getattribute__(self, name)
        if type(attr) == types.MethodType and name.startswith("req"):
            attr = _rate_limit_wrapper(attr)
            return attr
        else:
            return attr

    @staticmethod
    def _build_table_writers() -> Dict[str, TableWriter]:
        # noinspection PyDictCreation
        table_writers = {}

        ####
        # General
        ####

        table_writers["requests"] = TableWriter(["RequestId", "RequestType", *logger_contract.names(), "Note"],
                                                [dtypes.int64, dtypes.string, *logger_contract.types(), dtypes.string])

        table_writers["errors"] = TableWriter(
            ["RequestId", "ErrorCode", "ErrorDescription", "Error", "Note"],
            [dtypes.int64, dtypes.int64, dtypes.string, dtypes.string, dtypes.string])

        ####
        # Contracts
        ####

        table_writers["contracts_details"] = TableWriter(
            ["RequestId", *logger_contract_details.names()],
            [dtypes.int64, *logger_contract_details.types()])

        table_writers["contracts_matching"] = TableWriter(
            ["RequestId", *logger_contract.names(), "DerivativeSecTypes"],
            [dtypes.int64, *logger_contract.types(), dtypes.StringSet])

        table_writers["market_rules"] = TableWriter(
            ["MarketRuleId", *logger_price_increment.names()],
            [dtypes.string, *logger_price_increment.types()])

        ####
        # Accounts
        ####

        table_writers["accounts_managed"] = TableWriter(["Account"], [dtypes.string])

        table_writers["accounts_family_codes"] = TableWriter(
            [*logger_family_code.names()],
            [*logger_family_code.types()])

        table_writers["accounts_groups"] = TableWriter(
            ["GroupName", "DefaultMethod", "Account"],
            [dtypes.string, dtypes.string, dtypes.string])

        table_writers["accounts_allocation_profiles"] = TableWriter(
            ["AllocationProfileName", "Type", "Account", "Amount"],
            [dtypes.string, dtypes.string, dtypes.string, dtypes.float64])

        table_writers["accounts_aliases"] = TableWriter(
            ["Account", "Alias"],
            [dtypes.string, dtypes.string])

        table_writers["accounts_overview"] = TableWriter(
            ["RequestId", "Account", "ModelCode", "Currency", "Key", "Value"],
            [dtypes.int64, dtypes.string, dtypes.string, dtypes.string, dtypes.string, dtypes.string])

        table_writers["accounts_summary"] = TableWriter(
            ["RequestId", "Account", "Tag", "Value", "Currency"],
            [dtypes.int64, dtypes.string, dtypes.string, dtypes.string, dtypes.string])

        table_writers["accounts_positions"] = TableWriter(
            ["RequestId", "Account", "ModelCode", *logger_contract.names(), "Position", "AvgCost"],
            [dtypes.int64, dtypes.string, dtypes.string, *logger_contract.types(), dtypes.float64, dtypes.float64])

        table_writers["accounts_pnl"] = TableWriter(
            ["RequestId", "DailyPnl", "UnrealizedPnl", "RealizedPnl"],
            [dtypes.int64, dtypes.float64, dtypes.float64, dtypes.float64])
        
        table_writers["accounts_pnl_single"] = TableWriter(
            ["RequestId", "Position", "DailyPnL", "UnrealizedPnL", "RealizedPnL", "Value"],
            [dtypes.int64, dtypes.float64, dtypes.float64, dtypes.float64, dtypes.float64, dtypes.float64])

        ####
        # News
        ####

        table_writers["news_providers"] = TableWriter(logger_news_provider.names(), logger_news_provider.types())

        table_writers["news_bulletins"] = TableWriter(
            ["MsgId", "MsgType", "Message", "OriginExch"],
            [dtypes.int64, dtypes.string, dtypes.string, dtypes.string])

        table_writers["news_articles"] = TableWriter(
            ["RequestId", "ArticleType", "ArticleText"],
            [dtypes.int64, dtypes.string, dtypes.string])

        table_writers["news_historical"] = TableWriter(
            ["RequestId", "Timestamp", "ProviderCode", "ArticleId", "Headline"],
            [dtypes.int64, dtypes.Instant, dtypes.string, dtypes.string, dtypes.string])

        ####
        # Market Data
        ####

        table_writers["ticks_price"] = TableWriter(
            ["RequestId", "TickType", "Price", *logger_tick_attrib.names()],
            [dtypes.int64, dtypes.string, dtypes.float64, *logger_tick_attrib.types()])

        table_writers["ticks_size"] = TableWriter(
            ["RequestId", "TickType", "Size"],
            [dtypes.int64, dtypes.string, dtypes.float64])

        table_writers["ticks_string"] = TableWriter(
            ["RequestId", "TickType", "Value"],
            [dtypes.int64, dtypes.string, dtypes.string])

        # exchange for physical
        table_writers["ticks_efp"] = TableWriter(
            ["RequestId", "TickType", "BasisPoints", "FormattedBasisPoints", "TotalDividends", "HoldDays",
             "FutureLastTradeDate", "DividendImpact", "DividendsToLastTradeDate"],
            [dtypes.int64, dtypes.string, dtypes.float64, dtypes.string, dtypes.float64, dtypes.int64,
             dtypes.string, dtypes.float64, dtypes.float64])

        table_writers["ticks_generic"] = TableWriter(
            ["RequestId", "TickType", "Value"],
            [dtypes.int64, dtypes.string, dtypes.float64])

        table_writers["ticks_option_computation"] = TableWriter(
            ["RequestId", "TickType", "TickAttrib", "ImpliedVol", "Delta", "OptPrice", "PvDividend", "Gamma",
             "Vega", "Theta", "UndPrice"],
            [dtypes.int64, dtypes.string, dtypes.string, dtypes.float64, dtypes.float64, dtypes.float64, dtypes.float64, dtypes.float64,
             dtypes.float64, dtypes.float64, dtypes.float64])

        table_writers["ticks_trade"] = TableWriter(
            ["RequestId", *logger_hist_tick_last.names()],
            [dtypes.int64, *logger_hist_tick_last.types()])

        table_writers["ticks_bid_ask"] = TableWriter(
            ["RequestId", *logger_hist_tick_bid_ask.names()],
            [dtypes.int64, *logger_hist_tick_bid_ask.types()])

        table_writers["ticks_mid_point"] = TableWriter(
            ["RequestId", "Timestamp", "MidPoint"],
            [dtypes.int64, dtypes.Instant, dtypes.float64])

        table_writers["bars_historical"] = TableWriter(
            ["RequestId", *logger_bar_data.names()],
            [dtypes.int64, *logger_bar_data.types()])

        table_writers["bars_realtime"] = TableWriter(
            ["RequestId", *logger_real_time_bar_data.names()],
            [dtypes.int64, *logger_real_time_bar_data.types()])

        ####
        # Order Management System (OMS)
        ####

        table_writers["orders_submitted"] = TableWriter(
            [*logger_contract.names(), *logger_order.names(), *logger_order_state.names()],
            [*logger_contract.types(), *logger_order.types(), *logger_order_state.types()])

        table_writers["orders_status"] = TableWriter(
            ["OrderId", "Status", "Filled", "Remaining", "AvgFillPrice", "PermId", "ParentId", "LastFillPrice",
             "ClientId", "WhyHeld", "MktCapPrice"],
            [dtypes.int64, dtypes.string, dtypes.float64, dtypes.float64, dtypes.float64, dtypes.int64, dtypes.int64, dtypes.float64,
             dtypes.int64, dtypes.string, dtypes.float64])

        table_writers["orders_completed"] = TableWriter(
            [*logger_contract.names(), *logger_order.names(), *logger_order_state.names()],
            [*logger_contract.types(), *logger_order.types(), *logger_order_state.types()])

        table_writers["orders_exec_details"] = TableWriter(
            ["RequestId", *logger_contract.names(renames={"Exchange": "ContractExchange"}),
             *logger_execution.names(renames={"Exchange": "ExecutionExchange"})],
            [dtypes.int64, *logger_contract.types(), *logger_execution.types()])

        table_writers["orders_exec_commission_report"] = TableWriter(
            [*logger_commission_report.names()],
            [*logger_commission_report.types()])

        ####
        # End
        ####

        return table_writers

    ####################################################################################################################
    ####################################################################################################################
    ## Connect / Disconnect / Subscribe
    ####################################################################################################################
    ####################################################################################################################

    def connect(self, host: str, port: int, client_id: int) -> None:
        """Connect to an IB TWS session.  Raises an exception if already connected.

        Args:
            host (str): The host name or IP address of the machine where TWS is running. Leave blank to connect to the local host.
            port (int): TWS port, specified in TWS on the Configure>API>Socket Port field.
                By default production trading uses port 7496 and paper trading uses port 7497.
            client_id (int): A number used to identify this client connection.
                All orders placed/modified from this client will be associated with this client identifier.

                Note: Each client MUST connect with a unique clientId.

        Returns:
              None

        Raises:
              Exception
        """

        if self.isConnected():
            raise Exception("IbTwsClient is already connected.")

        self.contract_registry = ContractRegistry(self)
        self.order_id_queue = OrderIdEventQueue(self, strategy=self._order_id_strategy)
        self._registered_market_rules = set()
        self._realtime_bar_sizes = {}
        self.news_providers = []
        self._accounts_managed = set()

        EClient.connect(self, host, port, client_id)

        # wait for the client to connect to avoid a race condition (https://github.com/deephaven-examples/deephaven-ib/issues/12)
        time.sleep(1)

        self._thread = Thread(name="IbTwsClient", target=self.run)
        self._thread.start()
        setattr(self, "ib_thread", self._thread)

        # wait for the client to connect to avoid a race condition (https://github.com/deephaven-examples/deephaven-ib/issues/12)
        time.sleep(1)

        self._subscribe()

        # wait for the client to connect to avoid a race condition (https://github.com/deephaven-examples/deephaven-ib/issues/12)
        time.sleep(1)

    def disconnect(self) -> None:
        """Disconnect from an IB TWS session.

        Returns:
            None
        """

        EClient.disconnect(self)
        self._thread = None
        self.contract_registry = None
        self.order_id_queue = None
        self._registered_market_rules = None
        self._realtime_bar_sizes = None
        self.news_providers = None
        self._accounts_managed = None

    def __del__(self):
        self.disconnect()

    def _subscribe(self) -> None:
        """Subscribe to IB data."""

        self.reqFamilyCodes()

        if self._is_fa:
            self.requestFA(1)  # request GROUPS.  See FaDataTypeEnum.
            #TODO: see https://github.com/deephaven-examples/deephaven-ib/issues/32
            #TODO: see https://github.com/deephaven-examples/deephaven-ib/issues/5
            # self.requestFA(2)  # request PROFILE.  See FaDataTypeEnum.
            self.requestFA(3)  # request ACCOUNT ALIASES.  See FaDataTypeEnum.

            self.request_account_pnl("All")

        self.request_account_summary("All")
        self.request_account_overview("All")
        self.request_account_positions("All")
        self.reqManagedAccts()
        self.reqNewsBulletins(allMsgs=True)
        self.request_executions()
        self.reqNewsProviders()

        if not self._read_only:
            self.reqCompletedOrders(apiOnly=False)
            # Just subscribe to orders from the current client id.
            # When subscribing to all clients, data from other clients does not update.
            self.reqOpenOrders()

    ####################################################################################################################
    ####################################################################################################################
    ## General
    ####################################################################################################################
    ####################################################################################################################

    def log_request(self, req_id: int, request_type: str, contract: Optional[Contract],
                    notes: Optional[Dict[str, Any]]):
        """Log a data request."""

        if notes is None:
            note_string = None
        else:
            note_string = json.dumps(dict({k: str(v) for (k, v) in notes.items()}.items()))

        self._table_writers["requests"].write_row([req_id, request_type, *logger_contract.vals(contract), note_string])

    ####
    # Always present
    ####

    def error(self, reqId: TickerId, errorCode: int, errorString: str, advancedOrderRejectJson: str = ""):
        EWrapper.error(self, reqId, errorCode, errorString, advancedOrderRejectJson)

        if reqId == 2147483647:
            reqId = None

        if errorCode not in _error_code_message_map:
            msg = f"Unmapped error code.  Please file an issue at https://github.com/deephaven-examples/deephaven-ib/issues:\n\terrorCode='{errorCode}'\n\terrorString='{errorString}'\n\tThis only impacts the error message you see and will not impact the execution of your program."
            logging.error(msg)
            _error_code_message_map[errorCode] = errorString
            _error_code_note_map[errorCode] = ""

        self._table_writers["errors"].write_row(
            [reqId, errorCode, map_values(errorCode, _error_code_message_map), errorString,
             map_values(errorCode, _error_code_note_map)])

        # error may get called after disconnect, so need to avoid cases where contract_registry is None
        if self.isConnected():
            self.contract_registry.add_error_data(req_id=reqId, error_string=errorString)

    ####################################################################################################################
    ####################################################################################################################
    ## Contracts
    ####################################################################################################################
    ####################################################################################################################

    def request_market_rules(self, contractDetails: ContractDetails):
        """Request price increment market quoting rules, if they have not yet been retrieved."""

        for market_rule in contractDetails.marketRuleIds.split(","):
            if market_rule not in self._registered_market_rules:
                self.reqMarketRule(marketRuleId=int(market_rule))

    ####
    # reqContractDetails
    ####

    def contractDetails(self, reqId: int, contractDetails: ContractDetails):
        EWrapper.contractDetails(self, reqId, contractDetails)
        self._table_writers["contracts_details"].write_row([reqId, *logger_contract_details.vals(contractDetails)])
        self.contract_registry.add_contract_data(reqId, contractDetails)
        self.request_market_rules(contractDetails)

    def bondContractDetails(self, reqId: int, contractDetails: ContractDetails):
        EWrapper.bondContractDetails(self, reqId, contractDetails)
        self._table_writers["contracts_details"].write_row([reqId, *logger_contract_details.vals(contractDetails)])
        self.contract_registry.add_contract_data(reqId, contractDetails)
        self.request_market_rules(contractDetails)

    def contractDetailsEnd(self, reqId: int):
        EWrapper.contractDetailsEnd(self, reqId)
        self.contract_registry.request_end(reqId)

    ####
    # reqMatchingSymbols
    ####

    def symbolSamples(self, reqId: int, contractDescriptions: ListOfContractDescription):
        EWrapper.symbolSamples(self, reqId, contractDescriptions)

        for cd in contractDescriptions:
            self._table_writers["contracts_matching"].write_row([reqId, *logger_contract.vals(cd.contract),
                                                                 to_string_set(cd.derivativeSecTypes)])

            # Negative contract IDs seem to be for malformed contracts that yield errors when requesting details
            if cd.contract.conId >= 0:
                self.contract_registry.request_contract_details_nonblocking(cd.contract)

    ####
    # reqMarketRule
    ####

    def marketRule(self, marketRuleId: int, priceIncrements: ListOfPriceIncrements):
        EWrapper.marketRule(self, marketRuleId, priceIncrements)

        for pi in priceIncrements:
            self._table_writers["market_rules"].write_row([str(marketRuleId), *logger_price_increment.vals(pi)])

        self._registered_market_rules.add(str(marketRuleId))

    ####################################################################################################################
    ####################################################################################################################
    ## Accounts
    ####################################################################################################################
    ####################################################################################################################

    def request_account_summary(self, group_name: str) -> None:
        """Request account summary data for an account group."""

        account_summary_tags = [
            "accountountType",
            "NetLiquidation",
            "TotalCashValue",
            "SettledCash",
            "TotalCashValue",
            "AccruedCash",
            "BuyingPower",
            "EquityWithLoanValue",
            "PreviousDayEquityWithLoanValue",
            "GrossPositionValue",
            "RegTEquity",
            "RegTMargin",
            "SMA",
            "InitMarginReq",
            "MaintMarginReq",
            "AvailableFunds",
            "ExcessLiquidity",
            "Cushion",
            "FullInitMarginReq",
            "FullMaintMarginReq",
            "FullAvailableFunds",
            "FullExcessLiquidity",
            "LookAheadNextChange",
            "LookAheadInitMarginReq",
            "LookAheadMaintMarginReq",
            "LookAheadAvailableFunds",
            "LookAheadExcessLiquidity",
            "HighestSeverity",
            "DayTradesRemaining",
            "Leverage",
            "$LEDGER",
        ]

        req_id = self.request_id_manager.next_id()
        tags = ",".join(account_summary_tags)
        self.log_request(req_id, "AccountSummary", None, {"groupName": group_name, "tags": tags})
        self.reqAccountSummary(reqId=req_id, groupName=group_name, tags=tags)

    def request_account_pnl(self, account: str, model_code: str = "") -> int:
        """Request PNL updates.  Results are returned in the `accounts_pnl` table.

        Args:
            account (str): Account to request PNL for.  "All" requests for all accounts.
            model_code (str): Model portfolio code to request PNL for.

        Returns:
            Request ID

        Raises:
              Exception
        """

        req_id = self.request_id_manager.next_id()
        self.log_request(req_id, "Pnl", None, {"account": account, "model_code": model_code})
        self.reqPnL(reqId=req_id, account=account, modelCode=model_code)
        return req_id

    def request_account_overview(self, account: str, model_code: str = "") -> int:
        """Request portfolio overview updates.  Results are returned in the `accounts_overview` table.

        Args:
            account (str): Account to request an overview for.  "All" requests for all accounts.
            model_code (str): Model portfolio code to request an overview for.

        Returns:
            Request ID

        Raises:
              Exception
        """
        req_id = self.request_id_manager.next_id()
        self.log_request(req_id, "AccountOverview", None, {"account": account, "model_code": model_code})
        self.reqAccountUpdatesMulti(reqId=req_id, account=account, modelCode=model_code, ledgerAndNLV=False)
        return req_id

    def request_account_positions(self, account: str, model_code: str = "") -> int:
        """Request portfolio position updates.  Results are returned in the `accounts_positions` table.

        Args:
            account (str): Account to request positions for.  "All" requests for all accounts.
            model_code (str): Model portfolio code to request positions for.

        Returns:
            Request ID

        Raises:
              Exception
        """
        req_id = self.request_id_manager.next_id()
        self.log_request(req_id, "AccountPositions", None, {"account": account, "model_code": model_code})
        self.reqPositionsMulti(reqId=req_id, account=account, modelCode=model_code)
        return req_id

    def request_single_pnl(self, account: str, model_code: str, conid: int) -> int:
        """Request PNL updates for a single position.  Results are returned in the `accounts_pnl_single` table.

        Args:
            account (str): Account to request PNL for.
            model_code (str): Model portfolio code to request PNL for.
            con_id (int): Contract ID of the position to request PNL for.

        Returns:
            Request ID

        Raises:
              Exception
        """

        req_id = self.request_id_manager.next_id()
        self.log_request(req_id, "PnlSingle", None, {"account": account, "model_code": model_code, "conid": conid})
        self.reqPnLSingle(reqId=req_id, account=account, modelCode=model_code, conid=conid)
        return req_id

    ####
    # reqManagedAccts
    ####

    def managedAccounts(self, accountsList: str):
        EWrapper.managedAccounts(self, accountsList)

        for account in accountsList.split(","):
            if account and account not in self._accounts_managed:
                self._accounts_managed.add(account)
                self._table_writers["accounts_managed"].write_row([account])
                self.request_account_pnl(account)
                self.request_account_overview(account)
                self.request_account_positions(account)
                             

    ####
    # reqFamilyCodes
    ####

    def familyCodes(self, familyCodes: ListOfFamilyCode):
        EWrapper.familyCodes(self, familyCodes)

        for fc in familyCodes:
            self._table_writers["accounts_family_codes"].write_row(logger_family_code.vals(fc))

    ####
    # requestFA
    ####

    def receiveFA(self, faData: FaDataType, cxml: str):
        EWrapper.receiveFA(self, faData, cxml)

        fa_data_type = FaDataTypeEnum.to_str(faData)
        logging.debug(f"RECEIVEFA XML: {faData} {fa_data_type} {cxml}")

        xml_tree = ET.fromstring(cxml)

        if fa_data_type == "GROUPS":
            if xml_tree.tag != "ListOfGroups":
                raise Exception(f"Unexpected XML tag: {xml_tree.tag} != ListOfGroups")

            for group in xml_tree.findall("Group"):
                name = group.find("name").text
                accounts = group.find("ListOfAccts")
                default_method = group.find("defaultMethod").text

                for account in accounts.findall("Account"):
                    account = account.find("acct").text
                    self._table_writers["accounts_groups"].write_row([name, default_method, account])

                self.request_account_summary(name)

        elif fa_data_type == "PROFILES":
            if xml_tree.tag != "ListOfAllocationProfiles":
                raise Exception(f"Unexpected XML tag: {xml_tree.tag} != ListOfAllocationProfiles")

            for profile in xml_tree.findall("AllocationProfile"):
                name = profile.find("name").text
                ap_type = profile.find("type").text
                allocations = profile["ListOfAllocations"]

                for allocation in allocations.findall("Allocation"):
                    acct = allocation.find("acct").text
                    amount = allocation.find("amount").text
                    type_names = {1: "Percentages", 2: "Financial Ratios", 3: "Shares"}
                    self._table_writers["accounts_allocation_profiles"].write_row(
                        [name, map_values(ap_type, type_names), acct, float(amount)])

        elif fa_data_type == "ALIASES":
            if xml_tree.tag != "ListOfAccountAliases":
                raise Exception(f"Unexpected XML tag: {xml_tree.tag} != ListOfAccountAliases")

            for alias in xml_tree.findall("AccountAlias"):
                account = alias.find("account").text
                account_alias = alias.find("alias").text
                self._table_writers["accounts_aliases"].write_row([account, account_alias])

        else:
            logging.error(f"RECEIVEFA unknown data type: {faData} {fa_data_type} {cxml}")

    ####
    # reqAccountUpdates
    ####

    # NOT NEEDED. reqAccountUpdatesMulti is used instead.

    ####
    # reqAccountUpdatesMulti
    ####

    def accountUpdateMulti(self, reqId: int, account: str, modelCode: str, key: str, value: str, currency: str):
        EWrapper.accountUpdateMulti(self, reqId, account, modelCode, key, value, currency)
        self._table_writers["accounts_overview"].write_row([reqId, account, modelCode, currency, key, value])

    def accountUpdateMultiEnd(self, reqId: int):
        # do not need to implement
        EWrapper.accountUpdateMultiEnd(self, reqId)

    ####
    # reqAccountSummary
    ####

    def accountSummary(self, reqId: int, account: str, tag: str, value: str, currency: str):
        EWrapper.accountSummary(self, reqId, account, tag, value, currency)
        self._table_writers["accounts_summary"].write_row([reqId, account, tag, value, currency])

    ####
    # reqPnLSingle
    ####

    def pnlSingle(self, reqId: int, pos: decimal.Decimal, dailyPnL: float, unrealizedPnL: float, realizedPnL: float, value: float):
        EWrapper.pnlSingle(self, reqId, pos, dailyPnL, unrealizedPnL, realizedPnL, value)
        self._table_writers["accounts_pnl_single"].write_row([
            reqId, 
            pos, 
            dailyPnL, 
            unrealizedPnL, 
            realizedPnL, 
            value
        ])

    # reqPositions
    ####

    # NOT NEEDED. reqAccountUpdatesMulti is used instead.

    ####
    # reqPositionsMulti
    ####

    def positionMulti(self, reqId: int, account: str, modelCode: str, contract: Contract, pos: decimal.Decimal, avgCost: float):
        EWrapper.positionMulti(self, reqId, account, modelCode, contract, pos, avgCost)

        # The returned contract seems to be inconsistent with IB's API to request contract details.
        # This hack is to work around the problem.
        # See https://github.com/deephaven-examples/deephaven-ib/issues/33
        
        if contract.secType == "STK":
            contract.exchange = "SMART"

        self._table_writers["accounts_positions"].write_row(
            [reqId, account, modelCode, *logger_contract.vals(contract), pos, avgCost])

        self.contract_registry.request_contract_details_nonblocking(contract)

    def positionMultiEnd(self, reqId: int):
        # do not need to implement
        EWrapper.positionMultiEnd(self, reqId)

    ####
    # reqPnL
    ####

    def pnl(self, reqId: int, dailyPnL: float, unrealizedPnL: float, realizedPnL: float):
        EWrapper.pnl(self, reqId, dailyPnL, unrealizedPnL, realizedPnL)
        self._table_writers["accounts_pnl"].write_row([reqId, dailyPnL, unrealizedPnL, realizedPnL])

    ####################################################################################################################
    ####################################################################################################################
    ## News
    ####################################################################################################################
    ####################################################################################################################

    ####
    # reqNewsProviders
    ####

    def newsProviders(self, newsProviders: ListOfNewsProviders):
        EWrapper.newsProviders(self, newsProviders)

        for provider in newsProviders:
            self._table_writers["news_providers"].write_row(logger_news_provider.vals(provider))
            self.news_providers.append(provider.code)

    ####
    # reqNewsBulletins
    ####

    def updateNewsBulletin(self, msgId: int, msgType: int, newsMessage: str, originExch: str):
        EWrapper.updateNewsBulletin(self, msgId, msgType, newsMessage, originExch)
        self._table_writers["news_bulletins"].write_row([msgId, map_values(msgType, _news_msgtype_map), newsMessage,
                                                         originExch])

    ####
    # reqNewsArticle
    ####

    def newsArticle(self, requestId: int, articleType: int, articleText: str):
        EWrapper.newsArticle(self, requestId, articleType, articleText)
        at = map_values(articleType, {0: "PlainTextOrHtml", 1: "BinaryDataOrPdf"})
        self._table_writers["news_articles"].write_row([requestId, at, html.unescape(articleText)])

    ####
    # reqHistoricalNews
    ####

    def historicalNews(self, requestId: int, timestamp: str, providerCode: str, articleId: str, headline: str):
        EWrapper.historicalNews(self, requestId, timestamp, providerCode, articleId, headline)

        h = headline.split("}", 1)

        if len(h) == 1:
            headline_clean = h[0]
        else:
            headline_clean = h[1]

        self._table_writers["news_historical"].write_row(
            [requestId, ib_to_j_instant(timestamp), providerCode, articleId,
             headline_clean])

    def historicalNewsEnd(self, requestId: int, hasMore: bool):
        # do not need to implement
        EWrapper.historicalNewsEnd(self, requestId, hasMore)

    ####################################################################################################################
    ####################################################################################################################
    ## Market Data
    ####################################################################################################################
    ####################################################################################################################

    ####
    # reqMktData
    ####

    def tickPrice(self, reqId: TickerId, tickType: TickType, price: float, attrib: TickAttrib):
        EWrapper.tickPrice(self, reqId, tickType, price, attrib)

        if price == 0.0:
            price = None

        self._table_writers["ticks_price"].write_row([reqId, TickTypeEnum.to_str(tickType), price,
                                                      *logger_tick_attrib.vals(attrib)])

    def tickSize(self, reqId: TickerId, tickType: TickType, size: decimal.Decimal):
        EWrapper.tickSize(self, reqId, tickType, size)
        self._table_writers["ticks_size"].write_row([reqId, TickTypeEnum.to_str(tickType), size])

    def tickString(self, reqId: TickerId, tickType: TickType, value: str):
        EWrapper.tickString(self, reqId, tickType, value)
        self._table_writers["ticks_string"].write_row([reqId, TickTypeEnum.to_str(tickType), value])

    def tickEFP(self, reqId: TickerId, tickType: TickType, basisPoints: float,
                formattedBasisPoints: str, totalDividends: float,
                holdDays: int, futureLastTradeDate: str, dividendImpact: float,
                dividendsToLastTradeDate: float):
        EWrapper.tickEFP(self, reqId, tickType, basisPoints, formattedBasisPoints, totalDividends, holdDays,
                         futureLastTradeDate, dividendImpact, dividendsToLastTradeDate)
        self._table_writers["ticks_efp"].write_row(
            [reqId, TickTypeEnum.to_str(tickType), basisPoints, formattedBasisPoints,
             totalDividends, holdDays, futureLastTradeDate, dividendImpact,
             dividendsToLastTradeDate])

    def tickGeneric(self, reqId: TickerId, tickType: TickType, value: float):
        EWrapper.tickGeneric(self, reqId, tickType, value)
        self._table_writers["ticks_generic"].write_row([reqId, TickTypeEnum.to_str(tickType), value])

    def tickOptionComputation(self, reqId: TickerId, tickType: TickType, tickAttrib: int,
                              impliedVol: float, delta: float, optPrice: float, pvDividend: float,
                              gamma: float, vega: float, theta: float, undPrice: float):
        EWrapper.tickOptionComputation(self, reqId, tickType, tickAttrib, impliedVol, delta, optPrice, pvDividend,
                                       gamma, vega, theta, undPrice)
        ta = map_values(tickAttrib, {0: "Return-based", 1: "Price-based"})
        self._table_writers["ticks_option_computation"].write_row([reqId, TickTypeEnum.to_str(tickType), ta, impliedVol,
                                                                   delta,
                                                                   optPrice, pvDividend, gamma, vega, theta, undPrice])

    def tickSnapshotEnd(self, reqId: int):
        # do not ned to implement
        EWrapper.tickSnapshotEnd(self, reqId)

    ####
    # reqTickByTickData and reqHistoricalTicks
    ####

    def tickByTickAllLast(self, reqId: int, tickType: int, timestamp: int, price: float,
                          size: decimal.Decimal, tickAttribLast: TickAttribLast, exchange: str,
                          specialConditions: str):
        EWrapper.tickByTickAllLast(self, reqId, tickType, timestamp, price, size, tickAttribLast, exchange,
                                   specialConditions)

        t = HistoricalTickLast()
        t.time = timestamp
        t.tickAttribLast = tickAttribLast
        t.price = price
        t.size = size
        t.exchange = exchange
        t.specialConditions = specialConditions

        self._table_writers["ticks_trade"].write_row([reqId, *logger_hist_tick_last.vals(t)])

    # noinspection PyUnusedLocal
    def historicalTicksLast(self, reqId: int, ticks: ListOfHistoricalTickLast, done: bool):
        EWrapper.historicalTicksLast(self, reqId, ticks, done)

        for t in ticks:
            self._table_writers["ticks_trade"].write_row([reqId, *logger_hist_tick_last.vals(t)])

    def tickByTickBidAsk(self, reqId: int, timestamp: int, bidPrice: float, askPrice: float,
                         bidSize: decimal.Decimal, askSize: decimal.Decimal, tickAttribBidAsk: TickAttribBidAsk):
        EWrapper.tickByTickBidAsk(self, reqId, timestamp, bidPrice, askPrice, bidSize, askSize, tickAttribBidAsk)

        t = HistoricalTickBidAsk()
        t.time = timestamp
        t.tickAttribBidAsk = tickAttribBidAsk
        t.priceBid = bidPrice
        t.priceAsk = askPrice
        t.sizeBid = bidSize
        t.sizeAsk = askSize

        self._table_writers["ticks_bid_ask"].write_row([reqId, *logger_hist_tick_bid_ask.vals(t)])

    def historicalTicksBidAsk(self, reqId: int, ticks: ListOfHistoricalTickBidAsk, done: bool):

        for t in ticks:
            self._table_writers["ticks_bid_ask"].write_row([reqId, *logger_hist_tick_bid_ask.vals(t)])

    def tickByTickMidPoint(self, reqId: int, timestamp: int, midPoint: float):
        EWrapper.tickByTickMidPoint(self, reqId, timestamp, midPoint)
        self._table_writers["ticks_mid_point"].write_row([reqId, unix_sec_to_j_instant(timestamp), midPoint])

    def historicalTicks(self, reqId: int, ticks: ListOfHistoricalTick, done: bool):
        EWrapper.historicalTicks(self, reqId, ticks, done)

        for t in ticks:
            self._table_writers["ticks_mid_point"].write_row([reqId, unix_sec_to_j_instant(t.time), t.price])

    ####
    # reqHistoricalData
    ####

    def historicalData(self, reqId: int, bar: BarData):
        EWrapper.historicalData(self, reqId, bar)
        self._table_writers["bars_historical"].write_row([reqId, *logger_bar_data.vals(bar)])

    def historicalDataEnd(self, reqId: int, start: str, end: str):
        # do not ned to implement
        EWrapper.historicalDataEnd(self, reqId, start, end)

    def historicalDataUpdate(self, reqId: int, bar: BarData):
        EWrapper.historicalDataUpdate(self, reqId, bar)
        self._table_writers["bars_historical"].write_row([reqId, *logger_bar_data.vals(bar)])

    ####
    # reqRealTimeBars
    ####

    def reqRealTimeBars(self, reqId: TickerId, contract: Contract, barSize: int,
                        whatToShow: str, useRTH: bool,
                        realTimeBarsOptions: TagValueList):
        self._realtime_bar_sizes[reqId] = barSize
        EClient.reqRealTimeBars(self, reqId, contract, barSize, whatToShow, useRTH, realTimeBarsOptions)

    def realtimeBar(self, reqId: TickerId, timestamp: int, open_: float, high: float, low: float, close: float,
                    volume: decimal.Decimal, wap: decimal.Decimal, count: int):
        EWrapper.realtimeBar(self, reqId, timestamp, open_, high, low, close, volume, wap, count)
        bar_size = self._realtime_bar_sizes[reqId]
        bar = RealTimeBar(time=timestamp, endTime=timestamp + bar_size, open_=open_, high=high, low=low, close=close,
                          volume=volume,
                          wap=wap, count=count)
        self._table_writers["bars_realtime"].write_row([reqId, *logger_real_time_bar_data.vals(bar)])

    ####################################################################################################################
    ####################################################################################################################
    ## Order Management System (OMS)
    ####################################################################################################################
    ####################################################################################################################

    def next_order_id(self) -> int:
        """Gets the next valid order ID."""
        return self.request_id_manager.next_order_id(self.order_id_queue)

    def request_executions(self) -> None:
        """Requests executions."""
        req_id = self.request_id_manager.next_id()
        self.log_request(req_id, "Executions", None, None)
        self.reqExecutions(reqId=req_id, execFilter=ExecutionFilter())


    ####
    # reqIds
    ####

    def nextValidId(self, orderId: int):
        EWrapper.nextValidId(self, orderId)
        self.order_id_queue.add_value(orderId)

    ####
    # reqAllOpenOrders / reqOpenOrders
    ####

    def openOrder(self, orderId: OrderId, contract: Contract, order: Order, orderState: OrderState):
        EWrapper.openOrder(self, orderId, contract, order, orderState)

        if orderId != order.orderId:
            raise Exception("Order IDs do not match")

        self._table_writers["orders_submitted"].write_row(
            [*logger_contract.vals(contract), *logger_order.vals(order), *logger_order_state.vals(orderState)])
        self.contract_registry.request_contract_details_nonblocking(contract)

    def orderStatus(self, orderId: OrderId, status: str, filled: decimal.Decimal,
                    remaining: decimal.Decimal, avgFillPrice: float, permId: int,
                    parentId: int, lastFillPrice: float, clientId: int,
                    whyHeld: str, mktCapPrice: float):
        EWrapper.orderStatus(self, orderId, status, filled, remaining, avgFillPrice, permId, parentId, lastFillPrice,
                             clientId, whyHeld, mktCapPrice)
        self._table_writers["orders_status"].write_row(
            [orderId, status, filled, remaining, avgFillPrice, permId, parentId,
             lastFillPrice, clientId, whyHeld, mktCapPrice])

    def openOrderEnd(self):
        # do not ned to implement
        EWrapper.openOrderEnd(self)

    ####
    # reqCompletedOrders
    ####

    def completedOrder(self, contract: Contract, order: Order, orderState: OrderState):
        EWrapper.completedOrder(self, contract, order, orderState)
        self._table_writers["orders_completed"].write_row([*logger_contract.vals(contract), *logger_order.vals(order),
                                                           *logger_order_state.vals(orderState)])
        self.contract_registry.request_contract_details_nonblocking(contract)

    def completedOrdersEnd(self):
        # do not ned to implement
        EWrapper.completedOrdersEnd(self)

    ####
    # reqExecutions
    ####

    def execDetails(self, reqId: int, contract: Contract, execution: Execution):
        EWrapper.execDetails(self, reqId, contract, execution)
        self._table_writers["orders_exec_details"].write_row([reqId, *logger_contract.vals(contract),
                                                              *logger_execution.vals(execution)])
        self.contract_registry.request_contract_details_nonblocking(contract)

    def execDetailsEnd(self, reqId: int):
        # do not need to implement
        EWrapper.execDetailsEnd(self, reqId)

    def commissionReport(self, commissionReport: CommissionReport):
        EWrapper.commissionReport(self, commissionReport)
        self._table_writers["orders_exec_commission_report"].write_row(logger_commission_report.vals(commissionReport))

    ####################################################################################################################
    ####################################################################################################################
    ## End
    ####################################################################################################################
    ####################################################################################################################



"""Functionality for creating Deephaven tables."""

import logging
from typing import List, Any, Sequence, Optional, Set
import collections

from deephaven.time import dh_now
import jpy
from deephaven import DynamicTableWriter
from deephaven.table import Table
from deephaven import dtypes
from deephaven.dtypes import DType
import decimal
from decimal import Decimal

from .trace import trace_str


class TableWriter:
    """A writer for logging data to Deephaven dynamic tables.

    Empty strings are logged as None.
    """

    _dtw: DynamicTableWriter
    _string_indices: List[int]
    _receive_time: bool

    def __init__(self, names: List[str], types: List[DType], receive_time: bool = True):
        TableWriter._check_for_duplicate_names(names)
        self.names = names
        self.types = types
        self._receive_time = receive_time

        if receive_time:
            self.names.insert(0, "ReceiveTime")
            self.types.insert(0, dtypes.Instant)

        col_defs = {name: type for name, type in zip(names, types)}
        self._dtw = DynamicTableWriter(col_defs)
        self._string_indices = [i for (i, t) in enumerate(types) if t == dtypes.string]

    @staticmethod
    def _check_for_duplicate_names(names: List[str]) -> None:
        counts = collections.Counter(names)
        dups = [name for name, count in counts.items() if count > 1]

        if len(dups) > 0:
            raise Exception(f"Duplicate column names: {','.join(dups)}")

    def _check_logged_value_types(self, values: List) -> None:
        for n, t, v in zip(self.names, self.types, values):
            if v is None:
                continue

            if (t is dtypes.string and not isinstance(v, str)) or \
                    (t is dtypes.int64 and not isinstance(v, int)) or \
                    (t is dtypes.float64 and not isinstance(v, float)):
                logging.error(
                    f"TableWriter column type and value type are mismatched: column_name={n} column_type={t} value_type={type(v)} value={v}\n{trace_str()}\n-----")

    def table(self) -> Table:
        """Gets the table data is logged to."""
        return self._dtw.table

    def write_row(self, values: List) -> None:
        """Writes a row of data.  The input values may be modified."""

        if self._receive_time:
            values.insert(0, dh_now())

        for i in range(len(values)):
            if  isinstance(values[i], decimal.Decimal): 
                values[i] = float(values[i])

        self._check_logged_value_types(values)

        for i in self._string_indices:
            if values[i] == "":
                values[i] = None

        try:
            self._dtw.write_row(*values)
        except Exception as e:
            msg = f"Problem logging row:\n"

            for i, v in enumerate(values):
                msg += f"\t{i} {type(v)} {v}\n"

            logging.error(msg)

            raise e


ArrayStringSet = jpy.get_type("io.deephaven.stringset.ArrayStringSet")

_unmapped_values_already_logged:Set[str] = set()

def map_values(value, map, default=lambda v: f"UNKNOWN({v})") -> Any:
    """ Maps one set of values to another.  A default value is used if the value is not in the map. """

    if value is None:
        return None

    try:
        return map[value]
    except KeyError:
        msg = f"Unmapped value.  Please file an issue at https://github.com/deephaven-examples/deephaven-ib/issues: '{value}'\n{trace_str()}\n-----"

        if msg not in _unmapped_values_already_logged:
            _unmapped_values_already_logged.add(msg)
            logging.error(msg)

        return default(value)


def to_string_val(value) -> Optional[str]:
    """ Converts a value to a string. """

    if value is None:
        return None

    return str(value)


def to_string_set(value: Sequence) -> Optional[ArrayStringSet]:
    """ Converts an iterable to a string set. """

    if value is None:
        return None

    return ArrayStringSet(list({to_string_val(v) for v in value}))



from enum import Enum
from typing import Union, Dict, List, Callable, Optional
import json
import datetime
import numpy
import pandas

from deephaven.table import Table
from deephaven.constants import NULL_DOUBLE
from deephaven.dtypes import Instant
from ibapi.contract import Contract, ContractDetails
from ibapi.order import Order

from ._tws import IbTwsClient
from ._tws.order_id_queue import OrderIdStrategy
from .time import to_ib_datetime

__all__ = ["MarketDataType", "TickDataType", "BarDataType", "BarSize", "Duration", "OrderIdStrategy",
           "Request", "RegisteredContract", "IbSessionTws"]


class MarketDataType(Enum):
    """Type of market data to use."""

    REAL_TIME = 1
    """Real-time market data."""
    FROZEN = 2
    """Real-time market data during regular trading hours, and frozen prices after the close."""
    DELAYED = 3
    """Delayed market data."""


class TickDataType(Enum):
    """Tick data type."""

    LAST = "Last"
    """Most recent trade."""
    BID_ASK = "BidAsk"
    """"Most recent bid and ask."""
    MIDPOINT = "MidPoint"
    """Most recent midpoint."""

    def historical_value(self) -> str:
        if self.value == "Last":
            return "Trades"
        else:
            return self.value


class GenericTickType(Enum):
    """Tick data types for 'Generic' data.

    See: https://interactivebrokers.github.io/tws-api/tick_types.html
    """

    DIVIDENDS = 456
    """Dividends."""

    AUCTION = 225
    """Auction details."""

    MARK_PRICE = 232
    """Mark price is the current theoretical calculated value of an instrument. Since it is a calculated value, it will typically have many digits of precision."""
    MARK_PRICE_SLOW = 619
    """Slower mark price update used in system calculations."""

    TRADING_RANGE = 165
    """Multi-week price and volume trading ranges."""

    TRADE_LAST_RTH = 318
    """Last regular trading hours traded price."""
    TRADE_COUNT = 293
    """Trade count for the day."""
    TRADE_COUNT_RATE = 294
    """Trade count per minute."""
    TRADE_VOLUME = 233
    """Trade volume for the day."""
    TRADE_VOLUME_NO_UNREPORTABLE = 375
    """Trade volume for the day that excludes "Unreportable Trades"."""
    TRADE_VOLUME_RATE = 295
    """Trade volume per minute."""
    TRADE_VOLUME_SHORT_TERM = 595
    """Short-term trading volume."""

    SHORTABLE = 236
    """Describes the level of difficulty with which the contract can be sold short."""
    SHORTABLE_SHARES = 236
    """Number of shares available to short."""

    FUTURE_OPEN_INTEREST = 588
    """Total number of outstanding futures contracts."""
    FUTURE_INDEX_PREMIUM = 162
    """Number of points that the index is over the cash index."""

    DELAYED_YIELD_BID = 103
    """10-15 minute delayed bid yield price."""
    DELAYED_YIELD_ASK = 104
    """10-15 minute delayed ask yield price."""

    OPTION_VOLATILITY_HISTORICAL_REAL_TIME = 411
    """Real-time historical volatility."""
    OPTION_VOLATILITY_IMPLIED = 106
    """IB 30-day at-market volatility, estimated for a maturity thirty calendar days forward of the current trading day"""
    OPTION_VOLUME = 100
    """Option volume for the trading day."""
    OPTION_VOLUME_AVERAGE = 105
    """Average option volume for a trading day."""
    OPTION_OPEN_INTEREST = 101
    """Option open interest."""

    ETF_NAV_CLOSE = 578
    """ETF's Net Asset Value (NAV) closing price."""
    ETF_NAV_PRICE = 576
    """ETF's Net Asset Value (NAV) bid / ask price."""
    ETF_NAV_LAST = 577
    """ETF's Net Asset Value (NAV) last price."""
    ETF_NAV_LAST_FROZEN = 623
    """ETF's Net Asset Value (NAV) for frozen data."""
    ETF_NAV_RANGE = 614
    """ETF's Net Asset Value (NAV) price range."""

    BOND_FACTOR_MULTIPLIER = 460
    """Bond factor multiplier is a number that indicates the ratio of the current bond principal to the original principal."""


class BarDataType(Enum):
    """Bar data type."""

    TRADES = 1
    """Trade prices."""
    MIDPOINT = 2
    """Midpoint prices."""
    BID = 3
    """Bid prices."""
    ASK = 4
    """Ask prices."""
    BID_ASK = 5
    """Bid/Ask prices."""
    ADJUSTED_LAST = 6
    """Bid/Ask prices."""
    HISTORICAL_VOLATILITY = 7
    """Historical volatility."""
    OPTION_IMPLIED_VOLATILITY = 8
    """Option implied volatility."""
    REBATE_RATE = 9
    """Rebate rate."""
    FEE_RATE = 10
    """Fee rate."""
    YIELD_BID = 11
    """Bid yield."""
    YIELD_ASK = 12
    """Ask yield."""
    YIELD_BID_ASK = 13
    """Bid/Ask yield."""
    YIELD_LAST = 14
    """Last yield."""
    AGGTRADES = 15
    """Aggregate trade prices."""

class BarSize(Enum):
    """Bar data sizes."""

    SEC_1 = "1 sec"
    """1 second bar."""
    SEC_5 = "5 secs"
    """5 second bar."""
    SEC_10 = "10 secs"
    """10 second bar."""
    SEC_15 = "15 secs"
    """15 second bar."""
    SEC_30 = "30 secs"
    """30 second bar."""
    MIN_1 = "1 min"
    """1 minute bar."""
    MIN_2 = "2 mins"
    """2 minute bar."""
    MIN_3 = "3 mins"
    """3 minute bar."""
    MIN_5 = "5 mins"
    "5 minute bar."
    MIN_10 = "10 mins"
    "10 minute bar."
    MIN_15 = "15 mins"
    """15 minute bar."""
    MIN_20 = "20 mins"
    "20 minute bar."
    MIN_30 = "30 mins"
    """30 minute bar."""
    HOUR_1 = "1 hour"
    """1 hour bar."""
    HOUR_2 = "2 hour"
    """2 hour bar."""
    HOUR_3 = "3 hour"
    """3 hour bar."""
    HOUR_4 = "4 hour"
    """4 hour bar."""
    HOUR_8 = "8 hour"
    """8 hour bar."""
    DAY_1 = "1 day"
    """1 day bar."""
    WEEK_1 = "1W"
    """1 week bar."""
    MONTH_1 = "1M"
    """1 month bar."""


class Duration:
    """Time period to request data for."""

    value: str

    def __init__(self, value: str):
        self.value = value

    @staticmethod
    def seconds(value: int) -> "Duration":
        """Create a duration of a specified number of seconds.

        Args:
            value (int): number of seconds

        Returns:
            A duration.
        """
        return Duration(f"{value} S")

    @staticmethod
    def days(value: int) -> "Duration":
        """Create a duration of a specified number of days.

        Args:
            value (int): number of days

        Returns:
            A duration.
        """
        return Duration(f"{value} D")

    @staticmethod
    def weeks(value: int) -> "Duration":
        """Create a duration of a specified number of weeks.

        Args:
            value (int): number of weeks

        Returns:
            A duration.
        """
        return Duration(f"{value} W")

    @staticmethod
    def months(value: int) -> "Duration":
        """Create a duration of a specified number of months.

        Args:
            value (int): number of months

        Returns:
            A duration.
        """
        return Duration(f"{value} M")

    @staticmethod
    def years(value: int) -> "Duration":
        """Create a duration of a specified number of years.

        Args:
            value (int): number of years

        Returns:
            A duration.
        """
        return Duration(f"{value} Y")

    def __repr__(self) -> str:
        return f"Duration('{self.value}')"


class Request:
    """ IB session request. """

    request_id: int
    _cancel_func: Callable

    def __init__(self, request_id: int, cancel_func: Callable = None):
        self.request_id = request_id
        self._cancel_func = cancel_func

    def is_cancellable(self) -> bool:
        """Is the request cancellable?

        Returns:
            An indication if the request is cancellable.
        """
        return self._cancel_func is not None

    def cancel(self) -> None:
        """Cancel the request.

        Returns:
            None

        Raises:
            Exception: request is not cancellable.
        """

        if not self.is_cancellable():
            raise Exception("Request is not cancellable.")

        self._cancel_func(self.request_id)


class RegisteredContract:
    """ Details describing a financial instrument that has been registered in the framework.  This can be a stock, bond, option, etc.

    When some contracts are registered, details on multiple contracts are returned.
    """

    query_contract: Contract
    contract_details: List[ContractDetails]

    def __init__(self, query_contract: Contract, contract_details: List[ContractDetails]):
        self.query_contract = query_contract
        self.contract_details = contract_details

    def is_multi(self) -> bool:
        """Does the contract have multiple contract details?

        Returns:
            An indication if the requested contract is associated with multiple contract details.
        """
        return len(self.contract_details) > 1

    def __repr__(self) -> str:
        return f"RegistredContract({self.query_contract},[{'|'.join([str(cd.contract) for cd in self.contract_details])}])"


class IbSessionTws:
    """ IB TWS session.

    **NOTE: Some tables are data specific to the current client_id (e.g. orders_submitted).  A client_id of 0 includes
    data manually entered into the TWS session.  For example, orders entered by hand.**
    
    Args:
        host (str): The host name or IP address of the machine where TWS is running. Leave blank to connect to the local host.
            When run inside docker, you probably want ``host.docker.internal``.
        port (int): TWS port, specified in TWS on the ``Configure->API->Socket Port`` field.
            By default production trading uses port 7496 and paper trading uses port 7497.
        client_id (int): A number used to identify this client connection.
            All orders placed/modified from this client will be associated with this client identifier.

            **NOTE: Each client MUST connect with a unique clientId.**
        download_short_rates (bool): True to download a short rates table.
        order_id_strategy (OrderIdStrategy): strategy for obtaining new order ids.
        read_only (bool): True to create a read only client that can not trade; false to create a read-write client that can trade.  Default is true.
        is_fa (bool): True for financial advisor accounts; false otherwise.  Default is false.


    Tables:
        ####
        # General
        ####
        * **errors**: an error log.
        * **requests**: requests to IB.

        ####
        # Contracts
        ####
        * **contract_details**: details describing contracts of interest.  Automatically populated.
        * **contracts_matching**: contracts matching query strings provided to ``request_contracts_matching``.
        * **market_rules**: market rules indicating the price increment a contract can trade in.  Automatically populated.
        * **short_rates**: interest rates for shorting securities.  Automatically populated if ``download_short_rates=True``.


        ####
        # Accounts
        ####
        * **accounts_managed**: accounts managed by the TWS session login.  Automatically populated.
        * **accounts_family_codes**: account family.  Automatically populated.
        * **accounts_groups**: account groups.  Automatically populated.
        * **accounts_allocation_profiles**: allocation profiles for accounts.  Automatically populated.
        * **accounts_value**: account values.  Automatically populated.
        * **accounts_overview**: overview of account details.  Automatically populated.
        * **accounts_summary**: account summary.  Automatically populated.
        * **accounts_positions**: account positions.  Automatically populated.
        * **accounts_pnl**: account PNL.  Automatically populated.
        * **accounts_pnl_single**: single PNL.  populated by calling request_single_pnl() on a specific contract.
        
        ####
        # News
        ####

        * **news_providers**: currently subscribed news sources.  Automatically populated.
        * **news_bulletins**: news bulletins.  Automatically populated.
        * **news_articles**: the content of news articles requested via ``request_news_article``.
        * **news_historical**: historical news headlines requested via ``request_news_historical``.

        ####
        # Market Data
        ####

        * **ticks_price**: real-time tick market data of price values requested via ``request_market_data``.
        * **ticks_size**: real-time tick market data of size values requested via ``request_market_data``.
        * **ticks_string**: real-time tick market data of string values requested via ``request_market_data``.
        * **ticks_efp**: real-time tick market data of exchange for physical (EFP) values requested via ``request_market_data``.
        * **ticks_generic**: real-time tick market data of generic floating point values requested via ``request_market_data``.
        * **ticks_option_computation**: real-time tick market data of option computations requested via ``request_market_data``.
        * **ticks_trade**: real-time tick market data of trade prices requested via ``request_tick_data_historical`` or ``request_tick_data_realtime``.
        * **ticks_bid_ask**: real-time tick market data of bid and ask prices requested via ``request_tick_data_historical`` or ``request_tick_data_realtime``.
        * **ticks_mid_point**: real-time tick market data of mid-point prices requested via ``request_tick_data_historical`` or ``request_tick_data_realtime``.
        * **bars_historical**: historical price bars requested via ``request_bars_historical``.  Real-time bars change as new data arrives.
        * **bars_realtime**: real-time price bars requested via ``request_bars_realtime``.

        ####
        # Order Management System (OMS)
        ####

        * **orders_submitted**: submitted orders **FOR THE THE CLIENT ID**.  A client ID of 0 contains manually entered orders.  Automatically populated.
        * **orders_status**: order statuses.  Automatically populated.
        * **orders_completed**: completed orders.  Automatically populated.
        * **orders_exec_details**: order execution details.  Automatically populated.
        * **orders_exec_commission_report**: order execution commission report.  Automatically populated.
    """

    _host: str
    _port: int
    _client_id: int
    _read_only: bool
    _client: IbTwsClient
    _tables_raw: Dict[str, Table]
    _tables: Dict[str, Table]

    def __init__(self, host: str = "", port: int = 7497, client_id: int = 0, download_short_rates: bool = True, order_id_strategy: OrderIdStrategy = OrderIdStrategy.INCREMENT, read_only: bool = True, is_fa: bool = False):
        self._host = host
        self._port = port
        self._client_id = client_id
        self._read_only = read_only
        self._client = IbTwsClient(download_short_rates=download_short_rates, order_id_strategy=order_id_strategy, read_only=read_only, is_fa=is_fa)
        self._tables_raw = {f"raw_{k}": v for k, v in self._client.tables.items()}
        self._tables = dict(sorted(IbSessionTws._make_tables(self._tables_raw).items()))

    @property
    def host(self) -> str:
        """Client host.

        Returns:
            Client host.
        """
        return self._host

    @property
    def port(self) -> int:
        """Client port.

        Returns:
            Client port.
        """
        return self._port

    @property
    def client_id(self) -> int:
        """Client ID.

        Returns:
            Client ID.
        """
        return self._client_id

    @property
    def read_only(self) -> bool:
        """Is the client read only?

        Returns:
            a boolean indicating if the client is read only.
        """

    def __repr__(self) -> str:
        return f"IbSessionTws(host={self._host}, port={self._port}, client_id={self._client_id}, read_only={self._read_only})"

    ####################################################################################################################
    ####################################################################################################################
    ## Connect / Disconnect / Subscribe
    ####################################################################################################################
    ####################################################################################################################

    def connect(self) -> None:
        """Connect to an IB TWS session.  Raises an exception if already connected.

        Returns:
              None

        Raises:
              Exception: problem executing action.
        """

        self._client.connect(self._host, self._port, self._client_id)

    def disconnect(self) -> None:
        """Disconnect from an IB TWS session.

        Returns:
            None
        """

        self._client.disconnect()

    def is_connected(self) -> bool:
        """Is there a connection with TWS?

        Returns:
            an indication if the client is connected to TWS.
        """

        return self._client.isConnected()

    def _assert_connected(self) -> None:
        """Assert that the IbSessionTws is connected."""

        if not self.is_connected():
            raise Exception("IbSessionTws is not connected.")

    def _assert_read_write(self) -> None:
        """Assert that the IbSessionTws is read-write."""

        if self._read_only:
            raise Exception("IbSessionTws is read-only.  Set 'read_only=False' to enable read-write operations, such as trading.")

    ####################################################################################################################
    ####################################################################################################################
    ## General
    ####################################################################################################################
    ####################################################################################################################

    @staticmethod
    def _make_tables(tables_raw: Dict[str, Table]) -> Dict[str, Table]:
        def annotate_ticks(t):
            requests = tables_raw["raw_requests"] \
                .drop_columns(["ReceiveTime", "RequestType", "SecId", "SecIdType", "DeltaNeutralContract", "Note"])

            requests_col_names = [ c.name for c in requests.columns ]

            rst = t.natural_join(requests, on="RequestId").move_columns_up(requests_col_names)

            if "Timestamp" in [ c.name for c in rst.columns ]:
                if "TimestampEnd" in [ c.name for c in rst.columns ]:
                    rst = rst.move_columns_up(["RequestId", "ReceiveTime", "Timestamp", "TimestampEnd"])
                else:
                    rst = rst.move_columns_up(["RequestId", "ReceiveTime", "Timestamp"])
            else:
                rst = rst.move_columns_up(["RequestId", "ReceiveTime"])

            return rst

        def deephaven_ib_float_value(s: Optional[str]) -> Optional[float]:
            if not s:
                return NULL_DOUBLE

            try:
                return float(s)
            except ValueError:
                return NULL_DOUBLE

        def deephaven_ib_parse_note(note:str, key:str) -> Optional[str]:
            dict = json.loads(note)

            if key in dict:
                return dict[key]

            return None

        return {
            "requests": tables_raw["raw_requests"] \
                .move_columns_up(["RequestId", "ReceiveTime"]),
            "errors": tables_raw["raw_errors"] \
                .natural_join(tables_raw["raw_requests"] \
                             .drop_columns("Note").rename_columns("RequestTime=ReceiveTime"), on="RequestId") \
                .move_columns_up(["RequestId", "ReceiveTime"]),
            "contracts_details": tables_raw["raw_contracts_details"] \
                .move_columns_up(["RequestId", "ReceiveTime"]),
            "accounts_family_codes": tables_raw["raw_accounts_family_codes"] \
                .drop_columns("ReceiveTime"),
            "accounts_groups": tables_raw["raw_accounts_groups"] \
                .drop_columns("ReceiveTime"),
            "accounts_allocation_profiles": tables_raw["raw_accounts_allocation_profiles"] \
                .drop_columns("ReceiveTime"),
            "accounts_aliases": tables_raw["raw_accounts_aliases"] \
                .drop_columns("ReceiveTime"),
            "accounts_managed": tables_raw["raw_accounts_managed"] \
                .select_distinct("Account"),
            "accounts_positions": tables_raw["raw_accounts_positions"] \
                .last_by(["RequestId", "Account", "ModelCode", "ContractId"]) \
                .move_columns_up(["RequestId", "ReceiveTime"]),
            "accounts_overview": tables_raw["raw_accounts_overview"] \
                .last_by(["RequestId", "Account", "Currency", "Key"]) \
                .update("DoubleValue = (double)deephaven_ib_float_value(Value)") \
                .move_columns_up(["RequestId", "ReceiveTime"]),
            "accounts_summary": tables_raw["raw_accounts_summary"] \
                .natural_join(tables_raw["raw_requests"], on="RequestId", joins="Note") \
                .update("GroupName=(String)deephaven_ib_parse_note(Note,`groupName`)") \
                .drop_columns("Note") \
                .update("DoubleValue = (double)deephaven_ib_float_value(Value)") \
                .last_by(["RequestId", "GroupName", "Account", "Tag"]) \
                .move_columns_up(["RequestId", "ReceiveTime", "GroupName"]),
            "accounts_pnl": tables_raw["raw_accounts_pnl"] \
                .natural_join(tables_raw["raw_requests"], on="RequestId", joins="Note") \
                .update([
                    "Account=(String)deephaven_ib_parse_note(Note,`account`)",
                    "ModelCode=(String)deephaven_ib_parse_note(Note,`model_code`)"]) \
                .move_columns_up(["RequestId", "ReceiveTime", "Account", "ModelCode"]) \
                .drop_columns("Note") \
                .last_by("RequestId"),
            "accounts_pnl_single": tables_raw["raw_accounts_pnl_single"] \
                .natural_join(tables_raw["raw_requests"], on="RequestId", joins="Note") \
                .update([
                    "Account=(String)deephaven_ib_parse_note(Note,`account`)",
                    "ModelCode=(String)deephaven_ib_parse_note(Note,`model_code`)",
                    "ConId=(String)deephaven_ib_parse_note(Note,`conid`)"]) \
                .move_columns_up(["RequestId", "ReceiveTime", "Account", "ModelCode", "ConId"]) \
                .drop_columns("Note") \
                .last_by("RequestId"),
            "contracts_matching": tables_raw["raw_contracts_matching"] \
                .natural_join(tables_raw["raw_requests"], on="RequestId", joins="Pattern=Note") \
                .move_columns_up(["RequestId", "ReceiveTime", "Pattern"]) \
                .update("Pattern=(String)deephaven_ib_parse_note(Pattern,`pattern`)"),
            "market_rules": tables_raw["raw_market_rules"].select_distinct(["MarketRuleId", "LowEdge", "Increment"]),
            "news_bulletins": tables_raw["raw_news_bulletins"],
            "news_providers": tables_raw["raw_news_providers"] \
                .drop_columns("ReceiveTime"),
            "news_articles": tables_raw["raw_news_articles"] \
                .move_columns_up(["RequestId", "ReceiveTime"]),
            "news_historical": tables_raw["raw_news_historical"] \
                .natural_join(tables_raw["raw_requests"], on="RequestId", joins=["ContractId","SecType","Symbol","LocalSymbol"]) \
                .move_columns_up(["RequestId", "ReceiveTime", "Timestamp", "ContractId", "SecType", "Symbol",
                               "LocalSymbol"]),
            "orders_completed": tables_raw["raw_orders_completed"] \
                .move_columns_up(["ReceiveTime", "OrderId", "ClientId", "PermId", "ParentId"]),
            "orders_exec_commission_report": tables_raw["raw_orders_exec_commission_report"],
            "orders_exec_details": tables_raw["raw_orders_exec_details"] \
                .move_columns_up(["RequestId", "ReceiveTime", "Timestamp", "ExecId", "AcctNumber"]) \
                .rename_columns("Account=AcctNumber"),
            # The status on raw_orders_submitted is buggy, so using the status from raw_orders_status
            "orders_submitted": tables_raw["raw_orders_submitted"] \
                .last_by("PermId") \
                .drop_columns("Status") \
                .natural_join(tables_raw["raw_orders_status"].last_by("PermId"), on="PermId", joins="Status")
                .move_columns_up(["ReceiveTime", "Account", "ModelCode", "PermId", "ClientId", "OrderId", "ParentId",
                               "Status"]),
            "orders_status": tables_raw["raw_orders_status"] \
                .last_by("PermId") \
                .move_columns_up(["ReceiveTime", "PermId", "ClientId", "OrderId", "ParentId"]),
            "bars_historical": annotate_ticks(tables_raw["raw_bars_historical"]).last_by(["RequestId", "Timestamp", "ContractId"]),
            "bars_realtime": annotate_ticks(tables_raw["raw_bars_realtime"]),
            "ticks_efp": annotate_ticks(tables_raw["raw_ticks_efp"]),
            "ticks_generic": annotate_ticks(tables_raw["raw_ticks_generic"]),
            "ticks_mid_point": annotate_ticks(tables_raw["raw_ticks_mid_point"]),
            "ticks_option_computation": annotate_ticks(tables_raw["raw_ticks_option_computation"]),
            "ticks_price": annotate_ticks(tables_raw["raw_ticks_price"]),
            "ticks_size": annotate_ticks(tables_raw["raw_ticks_size"]),
            "ticks_string": annotate_ticks(tables_raw["raw_ticks_string"]),
            "ticks_trade": annotate_ticks(tables_raw["raw_ticks_trade"] \
                                          .rename_columns("TradeExchange=Exchange")),
            "ticks_bid_ask": annotate_ticks(tables_raw["raw_ticks_bid_ask"]),
        }


    @property
    def tables(self) -> Dict[str, Table]:
        """Gets a dictionary of all data tables.

        Returns:
            Dictionary of all data tables.
        """
        return self._tables

    @property
    def tables_raw(self) -> Dict[str, Table]:
        """Gets a dictionary of all raw data tables.  Raw tables are just as the data comes from IB.

        Returns:
            Dictionary of all raw data tables.
        """
        return self._tables_raw

    ####################################################################################################################
    ####################################################################################################################
    ## Contracts
    ####################################################################################################################
    ####################################################################################################################

    def get_registered_contract(self, contract: Contract) -> RegisteredContract:
        """Gets a contract that has been registered in the framework.  The registered contract is confirmed to
        exist in the IB system and contains a complete description of the contract.

        Args:
            contract (Contract): contract to search for

        Returns:
            A contract that has been registered with deephaven-ib.

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()

        try:
            cd = self._client.contract_registry.request_contract_details_blocking(contract)
            return RegisteredContract(query_contract=contract, contract_details=cd)
        except Exception as e:
            raise Exception(f"Error getting registered contract: contract={contract} {e}")

    def request_contracts_matching(self, pattern: str) -> Request:
        """Request contracts matching a pattern.  Results are returned in the ``contracts_matching`` table.

        Args:
            pattern (str): pattern to search for.  Can include part of a ticker or part of the company name.

        Returns:
            A Request.

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        req_id = self._client.request_id_manager.next_id()
        self._client.log_request(req_id, "MatchingSymbols", None, {"pattern": pattern})
        self._client.reqMatchingSymbols(reqId=req_id, pattern=pattern)
        return Request(request_id=req_id)

    ####################################################################################################################
    ####################################################################################################################
    ## Accounts
    ####################################################################################################################
    ####################################################################################################################

    def request_account_pnl(self, account: str, model_code: str = "") -> Request:
        """Request PNL updates.  Results are returned in the ``accounts_pnl`` table.

        Args:
            account (str): Account to request PNL for.
            model_code (str): Model portfolio code to request PNL for.

        Returns:
            A Request.

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        req_id = self._client.request_account_pnl(account, model_code)
        return Request(request_id=req_id)

    def request_account_overview(self, account: str, model_code: str = "") -> Request:
        """Request portfolio overview updates.  Results are returned in the ``accounts_overview`` table.

        Args:
            account (str): Account to request an overview for.  "All" requests for all accounts.
            model_code (str): Model portfolio code to request an overview for.

        Returns:
            A Request.

        Raises:
              Exception: problem executing action.
        """
        self._assert_connected()
        req_id = self._client.request_account_overview(account, model_code)
        return Request(request_id=req_id)

    def request_account_positions(self, account: str, model_code: str = "") -> Request:
        """Request portfolio position updates.  Results are returned in the ``accounts_positions`` table.

        Args:
            account (str): Account to request positions for.  "All" requests for all accounts.
            model_code (str): Model portfolio code to request positions for.

        Returns:
            A Request.

        Raises:
              Exception: problem executing action.
        """
        self._assert_connected()
        req_id = self._client.request_account_positions(account, model_code)
        return Request(request_id=req_id)

    def request_single_pnl(self, contract: RegisteredContract, account: str, model_code: str = "") -> Request:
        """Request PNL updates for a single position.  Results are returned in the ``accounts_pnl_single`` table.

        Args:
            contract (RegisteredContract): contract data is requested for.
            account (str): Account to request PNL for.
            model_code (str): Model portfolio code to request PNL for.

        Returns:
            A Request.

        Raises:
              Exception: problem executing action.
        """
        self._assert_connected()
        req_id = self._client.request_single_pnl(account, model_code, contract.contract_details[0].contract.conId)
        return Request(request_id=req_id)


    ####################################################################################################################
    ####################################################################################################################
    ## News
    ####################################################################################################################
    ####################################################################################################################

    def request_news_historical(self, contract: RegisteredContract,
                                start: Union[None, Instant, int, str, datetime.datetime, numpy.datetime64, pandas.Timestamp],
                                end: Union[None, Instant, int, str, datetime.datetime, numpy.datetime64, pandas.Timestamp],
                                provider_codes: List[str] = None, total_results: int = 100) -> List[Request]:
        """ Request historical news for a contract.  Results are returned in the ``news_historical`` table.

        Registered contracts that are associated with multiple contract details produce multiple requests.

        Args:
            contract (RegisteredContract): contract data is requested for
            provider_codes (List[str]): a list of provider codes.  By default, all subscribed codes are used.
            start (Union[None, Instant, int, str, datetime.datetime, numpy.datetime64, pandas.Timestamp]): marks the (exclusive) start of the date range.  See https://deephaven.io/core/pydoc/code/deephaven.time.html#deephaven.time.to_j_instant for supported inputs.
            end (Union[None, Instant, int, str, datetime.datetime, numpy.datetime64, pandas.Timestamp]): marks the (inclusive) end of the date range.  See https://deephaven.io/core/pydoc/code/deephaven.time.html#deephaven.time.to_j_instant for supported inputs.
            total_results (int): the maximum number of headlines to fetch (1 - 300)

        Returns:
            All of the requests created by the action.

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()

        if not provider_codes:
            provider_codes = self._client.news_providers

        pc = "+".join(provider_codes)
        requests = []

        for cd in contract.contract_details:
            req_id = self._client.request_id_manager.next_id()
            self._client.log_request(req_id, "HistoricalNews", cd.contract,
                                     {"provider_codes": provider_codes, "start": start, "end": end,
                                      "total_results": total_results})
            self._client.reqHistoricalNews(reqId=req_id, conId=cd.contract.conId, providerCodes=pc,
                                           startDateTime=to_ib_datetime(start, sub_sec=False),
                                           endDateTime=to_ib_datetime(end, sub_sec=False),
                                           totalResults=total_results, historicalNewsOptions=[])
            requests.append(Request(request_id=req_id))

        return requests

    def request_news_article(self, provider_code: str, article_id: str) -> Request:
        """ Request the text of a news article.  Results are returned in the ``news_articles`` table.

        Args:
            provider_code (str): short code indicating news provider, e.g. FLY
            article_id (str): id of the specific article

        Returns:
            A Request.

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        req_id = self._client.request_id_manager.next_id()
        self._client.log_request(req_id, "NewsArticle", None,
                                 {f"provider_code": provider_code, "article_id": article_id})
        self._client.reqNewsArticle(reqId=req_id, providerCode=provider_code, articleId=article_id,
                                    newsArticleOptions=[])
        return Request(request_id=req_id)

    ####################################################################################################################
    ####################################################################################################################
    ## Market Data
    ####################################################################################################################
    ####################################################################################################################

    def set_market_data_type(self, market_data_type: MarketDataType) -> None:
        """Sets the default type of market data.

        Args:
            market_data_type (MarketDataType): market data type

        Returns:
            None

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        self._client.reqMarketDataType(marketDataType=market_data_type.value)

    # noinspection PyDefaultArgument
    def request_market_data(self, contract: RegisteredContract, generic_tick_types: List[GenericTickType] = [],
                            snapshot: bool = False, regulatory_snapshot: bool = False) -> List[Request]:
        """ Request market data for a contract.  Results are returned in the ``ticks_price``, ``ticks_size``,
        ``ticks_string``, ``ticks_efp``, ``ticks_generic``, and ``ticks_option_computation`` tables.

        Registered contracts that are associated with multiple contract details produce multiple requests.


        Args:
            contract (RegisteredContract): contract data is requested for
            generic_tick_types (List[GenericTickType]): generic tick types being requested
            snapshot (bool): True to return a single snapshot of Market data and have the market data subscription cancel.
                Do not enter any genericTicklist values if you use snapshots.
            regulatory_snapshot (bool): True to get a regulatory snapshot.  Requires the US Value Snapshot Bundle for stocks.

        Returns:
            A Request.

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        generic_tick_list = ",".join([str(x.value) for x in generic_tick_types])
        requests = []

        for cd in contract.contract_details:
            req_id = self._client.request_id_manager.next_id()
            self._client.log_request(req_id, "MarketData", cd.contract,
                                     {"generic_tick_types": generic_tick_types, "snapshot": snapshot,
                                      "regulatory_snapshot": regulatory_snapshot})
            self._client.reqMktData(reqId=req_id, contract=cd.contract,
                                    genericTickList=generic_tick_list, snapshot=snapshot,
                                    regulatorySnapshot=regulatory_snapshot, mktDataOptions=[])
            requests.append(Request(request_id=req_id, cancel_func=self._cancel_market_data))

        return requests

    def _cancel_market_data(self, req_id: int) -> None:
        """Cancel a market data request.

        Args:
            req_id (int): request id

        Returns:
            None

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        self._client.cancelMktData(reqId=req_id)

    def request_bars_historical(self, contract: RegisteredContract,
                                duration: Duration, bar_size: BarSize, bar_type: BarDataType,
                                end: Union[None, Instant, int, str, datetime.datetime, numpy.datetime64, pandas.Timestamp] = None,
                                market_data_type: MarketDataType = MarketDataType.FROZEN,
                                keep_up_to_date: bool = True) -> List[Request]:
        """Requests historical bars for a contract.  Results are returned in the ``bars_historical`` table.

        Registered contracts that are associated with multiple contract details produce multiple requests.

        Args:
            contract (RegisteredContract): contract data is requested for
            end (Union[None, Instant, int, str, datetime.datetime, numpy.datetime64, pandas.Timestamp]): Ending timestamp of the requested data.  See https://deephaven.io/core/pydoc/code/deephaven.time.html#deephaven.time.to_j_instant for supported inputs.
            duration (Duration): Duration of data being requested by the query.
            bar_size (BarSize): Size of the bars that will be returned.
            bar_type (BarDataType): Type of bars that will be returned.
            market_data_type (MarketDataType): Type of market data to return after the close.
            keep_up_to_date (bool): True to continuously update bars

        Returns:
            All of the requests created by this action.

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        requests = []

        for cd in contract.contract_details:
            req_id = self._client.request_id_manager.next_id()
            self._client.log_request(req_id, "HistoricalData", cd.contract,
                                     {
                                         "end": end,
                                         "duration": duration,
                                         "bar_size": bar_size,
                                         "bar_type": bar_type,
                                         "market_data_type": market_data_type,
                                         "keep_up_to_date": keep_up_to_date,
                                     })
            self._client.reqHistoricalData(reqId=req_id, contract=cd.contract,
                                           endDateTime=to_ib_datetime(end, sub_sec=False),
                                           durationStr=duration.value, barSizeSetting=bar_size.value,
                                           whatToShow=bar_type.name, useRTH=(market_data_type == MarketDataType.FROZEN),
                                           formatDate=2, keepUpToDate=keep_up_to_date, chartOptions=[])
            requests.append(Request(request_id=req_id))

        return requests

    def request_bars_realtime(self, contract: RegisteredContract, bar_type: BarDataType, bar_size: int = 5,
                              market_data_type: MarketDataType = MarketDataType.FROZEN) -> List[Request]:
        """Requests real time bars for a contract.  Results are returned in the ``bars_realtime`` table.

        Registered contracts that are associated with multiple contract details produce multiple requests.

        Args:
            contract (RegisteredContract): contract data is requested for
            bar_type (BarDataType): Type of bars that will be returned.
            bar_size (int): Bar size in seconds.
            market_data_type (MarketDataType): Type of market data to return after the close.

        Returns:
            All of the requests created by this action.

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        requests = []

        if bar_type not in [BarDataType.TRADES, BarDataType.AGGTRADES, BarDataType.MIDPOINT, BarDataType.BID, BarDataType.ASK]:
            raise Exception(f"Unsupported bar type: {bar_type}")

        for cd in contract.contract_details:
            req_id = self._client.request_id_manager.next_id()
            self._client.log_request(req_id, "RealTimeBars", cd.contract,
                                     {"bar_type": bar_type, "bar_size": bar_size, "market_data_type": market_data_type})
            self._client.reqRealTimeBars(reqId=req_id, contract=cd.contract, barSize=bar_size,
                                         whatToShow=bar_type.name, useRTH=(market_data_type == MarketDataType.FROZEN),
                                         realTimeBarsOptions=[])
            requests.append(Request(request_id=req_id, cancel_func=self._cancel_bars_realtime))

        return requests

    def _cancel_bars_realtime(self, req_id: int) -> None:
        """Cancel a real-time bar request.

        Args:
            req_id (int): request id

        Returns:
            None

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        self._client.cancelRealTimeBars(reqId=req_id)

    def request_tick_data_realtime(self, contract: RegisteredContract, tick_type: TickDataType,
                                   number_of_ticks: int = 0, ignore_size: bool = False) -> List[Request]:
        """Requests real-time tick-by-tick data.  Results are returned in the ``ticks_trade``, ``ticks_bid_ask``,
        and ``ticks_mid_point`` tables.

        Registered contracts that are associated with multiple contract details produce multiple requests.

        Args:
            contract (RegisteredContract): contract data is requested for
            tick_type (TickDataType): Type of market data to return.
            number_of_ticks (int): Number of historical ticks to request.
            ignore_size (bool): should size values be ignored.

        Returns:
            All of the requests created by this action.

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        requests = []

        for cd in contract.contract_details:
            req_id = self._client.request_id_manager.next_id()
            self._client.log_request(req_id, "TickByTickData", cd.contract,
                                     {"tick_type": tick_type, "number_of_ticks": number_of_ticks,
                                      "ignore_size": ignore_size})
            self._client.reqTickByTickData(reqId=req_id, contract=cd.contract,
                                           tickType=tick_type.value,
                                           numberOfTicks=number_of_ticks, ignoreSize=ignore_size)
            requests.append(Request(request_id=req_id, cancel_func=self._cancel_tick_data_realtime))

        return requests

    def _cancel_tick_data_realtime(self, req_id: int) -> None:
        """Cancel a real-time tick-by-tick data request.

        Args:
            req_id (int): request id

        Returns:
            None

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        self._client.cancelTickByTickData(reqId=req_id)

    def request_tick_data_historical(self, contract: RegisteredContract,
                                     tick_type: TickDataType, number_of_ticks: int,
                                     start: Union[None, Instant, int, str, datetime.datetime, numpy.datetime64, pandas.Timestamp] = None,
                                     end: Union[None, Instant, int, str, datetime.datetime, numpy.datetime64, pandas.Timestamp] = None,
                                     market_data_type: MarketDataType = MarketDataType.FROZEN,
                                     ignore_size: bool = False) -> List[Request]:
        """Requests historical tick-by-tick data. Results are returned in the ``ticks_trade``, ``ticks_bid_ask``,
        and ``ticks_mid_point`` tables.

        Registered contracts that are associated with multiple contract details produce multiple requests.

        Args:
            contract (RegisteredContract): contract data is requested for
            start (Union[None, Instant, int, str, datetime.datetime, numpy.datetime64, pandas.Timestamp]): marks the (exclusive) start of the date range.  See https://deephaven.io/core/pydoc/code/deephaven.time.html#deephaven.time.to_j_instant for supported inputs.
            end (Union[None, Instant, int, str, datetime.datetime, numpy.datetime64, pandas.Timestamp]): marks the (inclusive) end of the date range.  See https://deephaven.io/core/pydoc/code/deephaven.time.html#deephaven.time.to_j_instant for supported inputs.
            tick_type (TickDataType): Type of market data to return.
            number_of_ticks (int): Number of historical ticks to request.
            market_data_type (MarketDataType): Type of market data to return after the close.
            ignore_size (bool): should size values be ignored.

        Returns:
            All of the requests created by this action.

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        what_to_show = tick_type.historical_value()
        requests = []

        if tick_type not in [TickDataType.MIDPOINT, TickDataType.LAST]:
            raise Exception(f"Unsupported tick data type: {tick_type}")

        for cd in contract.contract_details:
            req_id = self._client.request_id_manager.next_id()
            self._client.log_request(req_id, "HistoricalTicks", cd.contract,
                                     {"start": start,
                                      "end": end,
                                      "tick_type": tick_type,
                                      "number_of_ticks": number_of_ticks,
                                      "market_data_type": market_data_type,
                                      "ignore_size": ignore_size,
                                      })
            self._client.reqHistoricalTicks(reqId=req_id, contract=cd.contract,
                                            startDateTime=to_ib_datetime(start, sub_sec=False),
                                            endDateTime=to_ib_datetime(end, sub_sec=False),
                                            numberOfTicks=number_of_ticks, whatToShow=what_to_show,
                                            useRth=market_data_type.value,
                                            ignoreSize=ignore_size, miscOptions=[])
            requests.append(Request(request_id=req_id))

        return requests

    ####################################################################################################################
    ####################################################################################################################
    ## Order Management System (OMS)
    ####################################################################################################################
    ####################################################################################################################

    def order_place(self, contract: RegisteredContract, order: Order) -> Request:
        """Places an order.

        Args:
            contract (RegisteredContract): contract to place an order on
            order (Order): order to place

        Returns:
            A Request.

        Raises:
              Exception: problem executing action.
        """
        self._assert_connected()
        self._assert_read_write()

        if contract.is_multi():
            raise Exception(
                f"RegisteredContracts with multiple contract details are not supported for orders: {contract}")

        if order.orderId == 0 or order.orderId is None:
            req_id = self._client.next_order_id()
            order.orderId = req_id
        else:
            req_id = order.orderId

        cd = contract.contract_details[0]
        self._client.log_request(req_id, "PlaceOrder", cd.contract, {"order": f"Order({order})"})
        self._client.placeOrder(req_id, cd.contract, order)
        return Request(request_id=req_id, cancel_func=self.order_cancel)

    def order_cancel(self, order_id: int) -> None:
        """Cancels an order.

        Args:
            order_id (int): order ID

        Returns:
            None

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        self._assert_read_write()
        self._client.cancelOrder(orderId=order_id, manualCancelOrderTime="")

    def order_cancel_all(self) -> None:
        """Cancel all open orders.

        Returns:
            None

        Raises:
              Exception: problem executing action.
        """

        self._assert_connected()
        self._assert_read_write()
        self._client.reqGlobalCancel()




"""An event queue for managing order ID requests."""

from threading import Event, Thread
from time import sleep
from enum import Enum
from typing import List, Callable, TYPE_CHECKING

from .._internal.threading import LoggingLock
from .._internal.trace import trace_all_threads_str

# Type hints on IbTwsClient cause a circular dependency.
# This conditional import plus a string-based annotation avoids the problem.
if TYPE_CHECKING:
    from .tws_client import IbTwsClient


class OrderIdStrategy(Enum):
    """Strategy used to obtain order IDs."""

    def __new__(cls, retry: bool, tws_request: bool):
        obj = object.__new__(cls)
        obj.retry = retry
        obj.tws_request = tws_request
        return obj

    INCREMENT = (False, False)
    """Use the initial order ID and increment the value upon every call.  This is fast, but it may fail for multiple, concurrent sessions."""
    BASIC = (False, True)
    """Request a new order ID from TWS every time one is needed."""
    RETRY = (True, True)
    """Request a new order ID from TWS every time one is needed.  Retry if TWS does not respond quickly.  TWS seems to have a bug where it does not always respond."""


class OrderIdRequest:
    """An order ID request."""

    _event: Event
    _getter: Callable[[], int]
    _value: int
    _lock: LoggingLock

    def __init__(self, event: Event, getter: Callable[[], int]):
        self._event = event
        self._getter = getter
        self._value = None
        self._lock = LoggingLock("OrderIdRequest")

    def get(self) -> int:
        """A blocking call to get the order ID."""

        time_out = 60.0
        event_happened = self._event.wait(time_out)

        if not event_happened:
            trace = trace_all_threads_str()
            msg = f"OrderIdRequest.get() timed out after {time_out} sec.  A possible deadlock or TWS bug was detected!  You may be able to avoid this problem by using a different OrderIdStrategy.  Please create an issue at https://github.com/deephaven-examples/deephaven-ib/issues containing this error message\n{trace}\n"
            raise Exception(msg)

        with self._lock:
            if self._value is None:
                self._value = self._getter()

            return self._value


class OrderIdEventQueue:
    """A thread-safe queue for requesting and getting order IDs."""

    _events: List[Event]
    _values: List[int]
    _lock: LoggingLock
    _strategy: OrderIdStrategy
    _last_value: int
    _request_thread: Thread

    def __init__(self, client: 'IbTwsClient', strategy: OrderIdStrategy):
        self._events = []
        self._values = []
        self._lock = LoggingLock("OrderIdEventQueue")
        self._client = client
        self._strategy = strategy
        self._last_value = None

        if strategy.retry:
            self._request_thread = Thread(name="OrderIdEventQueueRetry", target=self._retry, daemon=True)
            self._request_thread.start()

    def _retry(self):
        """Re-requests IDs if there is no response."""

        while True:
            for event in self._events:
                self._client.reqIds(-1)

            sleep(0.01)

    def request(self) -> OrderIdRequest:
        """Requests data from the queue."""

        event = Event()

        with self._lock:
            self._events.append(event)

        if self._strategy.tws_request:
            self._client.reqIds(-1)
        else:
            self._increment_value()

        return OrderIdRequest(event, self._get)

    def add_value(self, value: int) -> None:
        """Adds a new value to the queue."""

        with self._lock:
            # Upon startup, add_value is called, to set the initial value
            self._last_value = value

            # if is to filter out values requested by ibapi during initialization
            if self._events:
                self._values.append(value)
                event = self._events.pop(0)
                event.set()

    def _increment_value(self) -> None:
        """Increments the latest value and adds the value to the queue."""

        with self._lock:
            if self._events:
                self._values.append(self._last_value)
                self._last_value += 1
                event = self._events.pop(0)
                event.set()

    def _get(self) -> int:
        """Gets a value from the queue."""

        with self._lock:
            return self._values.pop(0)



"""A registry for managing contracts."""

import threading
from typing import TYPE_CHECKING

from ibapi.contract import Contract, ContractDetails

from .ib_type_logger import *
from .._internal.threading import LoggingLock

# Type hints on IbTwsClient cause a circular dependency.
# This conditional import plus a string-based annotation avoids the problem.
if TYPE_CHECKING:
    from .tws_client import IbTwsClient


class ContractEntry:
    """Entry in the ContractRegistry."""

    contract: Contract
    contract_details: List[ContractDetails]
    error_strings: List[str]

    def __init__(self, contract: Contract):
        self.contract = contract
        self.contract_details = []
        self.error_strings = []

    def add_contract_details(self, contract_details: ContractDetails):
        """Adds contract details to the entry."""

        if self.error_strings:
            raise Exception(f"Adding contract details to an entry that already has an error string: {self.contract}")

        self.contract_details.append(contract_details)

    def add_error_string(self, req_id: int, error_string: str):
        """Adds an error string to the entry."""

        if self.contract_details:
            raise Exception(f"Adding an error string to an entry that already has contract details: req_id={req_id} {self.contract}")

        self.error_strings.append(f"req_id={req_id} {error_string}")

    def get(self) -> List[ContractDetails]:
        """Gets the details or raises an exception if there are no details."""
        if self.contract_details:
            return self.contract_details
        elif self.error_strings:
            raise Exception("|".join(self.error_strings))
        else:
            raise Exception(f"Contract has no details and no error: {self.contract}")


class ContractRegistry:
    """A registry for mapping between contract requests and official contract specifications."""

    _client: 'IbTwsClient'
    _lock: LoggingLock
    _requests_by_id: Dict[int, Tuple[Contract, threading.Event]]
    _requests_by_key: Dict[str, Tuple[Contract, threading.Event]]
    _contracts: Dict[str, ContractEntry]


    def __init__(self, client: 'IbTwsClient'):
        self._client = client
        self._lock = LoggingLock("ContractRegistry")
        self._requests_by_id = {}
        self._requests_by_key = {}
        self._contracts = {}


    def add_contract_data(self, req_id: int, contract_details: ContractDetails) -> None:
        """Add new contract details.

        Args:
            req_id (int): Request ID
            contract_details (ContractDetails): Contract details

        Returns:
            None
        """

        with self._lock:
            if not req_id in self._requests_by_id:
                return

            (contract, event) = self._requests_by_id[req_id]
            self._update_contract_details(contract, contract_details)
            self._update_contract_details(contract_details.contract, contract_details)


    def add_error_data(self, req_id: int, error_string: str) -> None:
        """Add new error details.

        Args:
            req_id (int): Request ID
            error_string (str): error string

        Returns:
            None
        """

        with self._lock:
            if not req_id in self._requests_by_id:
                return

            contract, event = self._requests_by_id[req_id]
            self._update_error(contract, req_id, error_string)

            if event:
                event.set()

    def request_end(self, req_id: int) -> None:
        """Indicate that the request is over and all data has been received."""

        with self._lock:
            if not req_id in self._requests_by_id:
                return

            contract, event = self._requests_by_id.pop(req_id)

            if event:
                event.set()


    def request_contract_details_nonblocking(self, contract: Contract) -> None:
        """Request contract details, if they have not yet been retrieved.
        
        Function does not block waiting for results.

        Args:
            contract (Contract): Contract being queried.
            
        Returns:
            None
        """

        key = str(contract)

        with self._lock:
            if key in self._requests_by_key:
                return

        self._request_contract_details(contract=contract)

    def request_contract_details_blocking(self, contract: Contract) -> List[ContractDetails]:
        """Request contract details, if they have not yet been retrieved.
        
        Function blocks waiting for results.

        Args:
            contract (Contract): Contract being queried.
            
        Returns:
            ContractDetails

        Raises:
            Exception
        """

        cd = self._get_contract_details(contract)

        if cd is not None:
            return cd.get()
        else:
            key = str(contract)

            with self._lock:
                if key in self._requests_by_key:
                    _, event = self._requests_by_key[key]
                    new_request = False
                else:
                    event = threading.Event()
                    new_request = True

            if new_request:
                self._request_contract_details(contract=contract, event=event)

            time_out = 2 * 60.0
            event_happened = event.wait(time_out)

            if not event_happened:
                raise Exception(f"ContractRegistry.request_contract_details_blocking() timed out after {time_out} sec.  contract={contract}")

            cd = self._get_contract_details(contract)
            return cd.get()

    def _request_contract_details(self, contract: Contract, event: threading.Event = None) -> None:
        """Request contract details, if they have not yet been retrieved.

        Args:
            contract (Contract): Contract being queried.
            event (threading.Event): Event used to notify the requester that the contract details have been received.

        Returns:
            None
        """

        key = str(contract)

        with self._lock:
            if key not in self._contracts:
                if contract.conId < 0:
                    raise Exception("Requesting contract details for a contract with a negative conId.  This is almost certainly a bug.  Please submit a bug report with this stack trace: {contract}")

                req_id = self._client.request_id_manager.next_id()
                req = (contract, event)
                self._requests_by_id[req_id] = req
                self._requests_by_key[key] = req
                self._client.log_request(req_id, "ContractDetails", contract, None)
                self._client.reqContractDetails(reqId=req_id, contract=contract)

    def _get_contract_details(self, contract: Contract) -> ContractEntry:
        """Gets the contract details for a query contract."""

        key = str(contract)

        with self._lock:
            return self._contracts.get(key, None)

    def _update_contract_details(self, contract: Contract, contract_details: ContractDetails) -> None:
        """Updates the contract details for a query contract."""

        key = str(contract)

        if key not in self._contracts:
            self._contracts[key] = ContractEntry(contract)

        self._contracts[key].add_contract_details(contract_details)

    def _update_error(self, contract: Contract, req_id: int, error_string: str) -> None:
        """Updates the error string for a query contract."""

        key = str(contract)

        if key not in self._contracts:
            self._contracts[key] = ContractEntry(contract)

        self._contracts[key].add_error_string(req_id, error_string)



"""Functionality for working with time in Deephaven and IB."""

from typing import Union

import jpy
import datetime
import numpy
import pandas

import deephaven.time as dtime
from deephaven.dtypes import Instant

_SimpleDateFormat = jpy.get_type("java.text.SimpleDateFormat")
_TimeZone = jpy.get_type("java.util.TimeZone")
_DateTimeUtils = jpy.get_type("io.deephaven.time.DateTimeUtils")

_ib_date_time_pattern_sec = "yyyyMMdd HH:mm:ss"
_ib_date_time_pattern_subsec = "yyyyMMdd HH:mm:ss.S"
_ib_date_time_patterns = [
    "yyyyMMdd HH:mm:ss.S",
    "yyyy-MM-dd HH:mm:ss.S",
    "yyyyMMdd HH:mm:ss",
    "yyyy-MM-dd HH:mm:ss",
]
_ib_date_time_formatter_sec = _SimpleDateFormat(_ib_date_time_pattern_sec)
_ib_date_time_formatter_sec.setTimeZone(_TimeZone.getTimeZone("US/Eastern"))
_ib_date_time_formatter_subsec = _SimpleDateFormat(_ib_date_time_pattern_subsec)
_ib_date_time_formatter_subsec.setTimeZone(_TimeZone.getTimeZone("US/Eastern"))
_ib_date_time_formatters = [_SimpleDateFormat(pattern) for pattern in _ib_date_time_patterns]

for _f in _ib_date_time_formatters:
    _f.setTimeZone(_TimeZone.getTimeZone("US/Eastern"))


def to_ib_datetime(time: Union[None, Instant, int, str, datetime.datetime, numpy.datetime64, pandas.Timestamp],
                   sub_sec: bool = True) -> str:
    """Convert a time to an IB timestamp string.

    Args:
        time (Union[None, Instant, int, str, datetime.datetime, numpy.datetime64, pandas.Timestamp]): time.  See https://deephaven.io/core/pydoc/code/deephaven.time.html#deephaven.time.to_j_instant for supported inputs.
        sub_sec (bool): true to return subsecond resolution and false otherwise.
    """

    time = dtime.to_j_instant(time)

    if time is None:
        return ""

    date = _DateTimeUtils.toDate(time)

    if sub_sec:
        return _ib_date_time_formatter_subsec.format(date) + " US/Eastern"
    else:
        return _ib_date_time_formatter_sec.format(date) + " US/Eastern"


def ib_to_j_instant(time: str) -> Instant:
    """Convert an IB timestamp to a Java Instant."""

    if time is None:
        return None

    exceptions = []

    for formatter in _ib_date_time_formatters:
        try:
            return formatter.parse(time).toInstant()
        except Exception as e:
            exceptions.append(e)
            pass

    raise Exception(f"Unable to parse time '{time}'", exceptions)


def unix_sec_to_j_instant(time: int) -> Instant:
    """Convert Unix seconds since the epoch to a Java Instant."""

    if time is None:
        return None

    return dtime.to_j_instant(int(time) * 1000000000)



"""Functionality for working with requests."""

from .order_id_queue import OrderIdEventQueue
from .._internal.threading import LoggingLock


class RequestIdManager:
    """A manager for getting unique request IDs that are well behaved."""

    _lock: LoggingLock
    _id: int

    def __init__(self):
        self._lock = LoggingLock("RequestManager")
        self._id = 0

    def next_id(self) -> int:
        """Gets the next sequential ID for a generic request."""
        with self._lock:
            self._id += 1
            return self._id

    def next_order_id(self, order_id_queue: OrderIdEventQueue) -> int:
        """Gets the next sequential ID for an order request."""
        with self._lock:
            request = order_id_queue.request()
            oid = request.get()
            max_id = max(oid, self._id + 1)
            self._id = max_id
            return max_id



"""Functionality for working with error codes."""

from typing import Dict, Tuple

import pandas as pd


def load_error_codes() -> Tuple[Dict[int, str], Dict[int, str]]:
    """Load dictionaries of error code messages and notes."""

    html_tables = pd.read_html('https://interactivebrokers.github.io/tws-api/message_codes.html')

    error_messages = {}
    error_notes = {}

    for df in html_tables:
        try:
            df = df.fillna('')
            codes = df['Code']
            messages = df['TWS message']
            notes = df['Additional notes']

            for code, message in zip(codes, messages):
                error_messages[code] = message

            for code, note in zip(codes, notes):
                error_notes[code] = note
        except KeyError:
            pass

    overrides = {
        0: "Warning: Approaching max rate of 50 messages per second",
        504: "Not connected",
        502: "Couldn't connect to TWS. Confirm that 'Enable ActiveX and Socket EClients' is enabled and connection port is the same as 'Socket Port' on the TWS 'Edit->Global Configuration...->API->Settings' menu. Live Trading ports: TWS: 7496; IB Gateway: 4001. Simulated Trading ports for new installations of version 954.1 or newer:  TWS: 7497; IB Gateway: 4002",
        2113: "The order size for Bonds (Bills) is entered as a nominal par value of the order, and must be a multiple",
        2129: "The product is provided on an indicative and informational basis only. IBKR does not represent that the valuation for this instrument is accurate. The basis for the calculation may change at any time. Traders are responsible for understanding the contract details and details of deliverable instruments independently of IBKR sources, which are provided on a best efforts basis only.",
        2157: "Sec-def data farm connection is broken:secdefil",
        10089: "Requested market data requires additional subscription for API.See link in 'Market Data Connections' dialog for more details.",
        10147: "OrderId that needs to be cancelled is not found.",
        10168: "Requested market data is not subscribed.Delayed market data is not enabled.",
        10172: "Failed to request news article: No data available",
        10187: "Failed to request historical ticks",
        10189: "Failed to request tick-by-tick data",
        10190: "Maxnumber of tick-by-tick requests has been reached.",
        10311: "This order will be directly routed to NYSE. Direct routed orders may result in higher trade fees. Restriction is specified in Precautionary Settings of Global Configuration/API.",
        10349: "Order TIF was set to DAY based on order preset.",
    }

    for k, v in overrides.items():
        if k not in error_messages:
            error_messages[k] = v
            error_notes[k] = ""

    return error_messages, error_notes



"""Functionality for working with short rates."""

import ftplib
import html
import tempfile

from deephaven import read_csv
from deephaven.table import Table


class IBFtpWriter:
    """Writer for downloading text pipe-separated-value files from the IB FTP site.

    Closing the writer causes the temporary file containing the data to be deleted.
    """

    header: str
    source: str
    file: tempfile.NamedTemporaryFile

    def __init__(self):
        self.header = None
        self.source = None
        self.file = tempfile.NamedTemporaryFile(mode="w", suffix=".psv")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def file_name(self) -> str:
        """Name of the temporary file."""
        return self.file.name

    def flush(self) -> None:
        """Flush writes to the temporary file."""
        self.file.flush()

    def close(self) -> None:
        """Close the temporary file.  This makes the temporary file unavailable."""
        self.file.close()

    def write(self, line: str) -> None:
        """Write a line to the temporary file."""

        if line.startswith("#BOF") or line.startswith("#EOF"):
            return

        line = html.unescape(line)

        # REMOVE TRAILING "|" that breaks CSV parser
        # https://github.com/deephaven/deephaven-core/issues/1800
        if line.endswith("|"):
            line = line[:-1]

        if line.startswith("#"):
            line = f"Source|{line[1:]}"

            if self.header is None:
                self.header = line
                self.file.write(f"{line}\n")
            elif self.header != line:
                raise Exception(f"Mismatched headers: {self.header} {line}")
            else:
                return
        else:
            self.file.write(f"{self.source}|{line}\n")


def load_short_rates() -> Table:
    """Downloads the short rates from the IB FTP site and returns them as a table."""

    # See: https://www.ibkrguides.com/kb/article-2024.htm
    host: str = "ftp2.interactivebrokers.com"
    user: str = "shortstock"

    with ftplib.FTP(host=host, user=user) as ftp, IBFtpWriter() as p:
        try:
            files = ftp.nlst("*.txt")

            for file in files:
                p.source = file[:-4]
                res = ftp.retrlines(f'RETR {file}', p.write)

                if not res.startswith('226 Transfer complete'):
                    raise Exception(f"FTP download failed: {user}@{host} {file} {res}")

        except ftplib.all_errors as e:
            print('FTP error:', e)

        p.flush()

        return read_csv(p.file_name(), delimiter="|") \
            .rename_columns([
                "Sym=SYM",
                "Currency=CUR",
                "Name=NAME",
                "Contract=CON",
                "RebateRate=REBATERATE",
                "FeeRate=FEERATE",
                "Available=AVAILABLE"])



"""Functionality for logging IB types to Deephaven tables."""

import sys
from typing import Any, List, Tuple, Dict, Callable, Optional

from deephaven import dtypes

from .._internal.tablewriter import map_values, to_string_val, to_string_set
from ..time import unix_sec_to_j_instant, ib_to_j_instant


class IbComplexTypeLogger:
    """ Base class for logging complex IB types. """

    ib_type: str
    column_details: List[Tuple[str, Any, Callable]]

    def __init__(self, ib_type: str, column_details: List[Tuple[str, Any, Callable]]):
        self.ib_type = ib_type
        self.column_details = column_details

    # noinspection PyDefaultArgument
    def names(self, renames: Dict[str, str] = {}) -> List[str]:
        """ Column names. """
        return [renames.get(cd[0], cd[0]) for cd in self.column_details]

    def types(self) -> List[Any]:
        """ Column types. """
        return [cd[1] for cd in self.column_details]

    def vals(self, ib_obj: Any) -> List[Any]:
        """ Column values extracted from the IB object. """

        if ib_obj is None:
            return [None] * len(self.column_details)

        return [cd[2](ib_obj) for cd in self.column_details]


###

def _include_details(details: List[Tuple], lambda_for_field: Callable) -> List[Tuple]:
    """Details for logging one type within another.

    Args:
        details (List[Tuple]): details for logging the included type.
        lambda_for_field (Callable): lambda for extracting the included value from the type.

    Returns:
        Details for logging the inner type.
    """

    # To understand the bound variable voodoo, see: https://stackoverflow.com/questions/19837486/lambda-in-a-loop
    return [(d[0], d[1], lambda xx, bound_d2=d[2]: bound_d2(lambda_for_field(xx))) for d in details]


####

def _details_family_code() -> List[Tuple]:
    """Details for logging FamilyCode."""

    return [
        ("AccountID", dtypes.string, lambda fc: fc.accountID),
        ("FamilyCode", dtypes.string, lambda fc: fc.familyCodeStr),
    ]


logger_family_code = IbComplexTypeLogger("FamilyCode", _details_family_code())


####

def _details_contract() -> List[Tuple]:
    """ Details for logging Contract. """

    def map_right(right: str) -> Optional[str]:
        if right == "?":
            return None

        return right

    def map_multiplier(multiplier: str) -> float:
        if multiplier is None or multiplier == "":
            return 1.0

        return float(multiplier)

    return [
        ("ContractId", dtypes.int64, lambda contract: contract.conId),
        ("SecId", dtypes.string, lambda contract: contract.secId),
        ("SecIdType", dtypes.string, lambda contract: contract.secIdType),
        ("SecType", dtypes.string, lambda contract: contract.secType),
        ("Symbol", dtypes.string, lambda contract: contract.symbol),
        ("LocalSymbol", dtypes.string, lambda contract: contract.localSymbol),
        ("TradingClass", dtypes.string, lambda contract: contract.tradingClass),
        ("Currency", dtypes.string, lambda contract: contract.currency),
        ("Exchange", dtypes.string, lambda contract: contract.exchange),
        ("PrimaryExchange", dtypes.string, lambda contract: contract.primaryExchange),
        ("LastTradeDateOrContractMonth", dtypes.string, lambda contract: contract.lastTradeDateOrContractMonth),
        ("Strike", dtypes.float64, lambda contract: float(contract.strike)),
        ("Right", dtypes.string, lambda contract: map_right(contract.right)),
        ("Multiplier", dtypes.float64, lambda contract: map_multiplier(contract.multiplier)),

        # combos
        ("ComboLegsDescrip", dtypes.string, lambda contract: contract.comboLegsDescrip),
        ("ComboLegs", dtypes.StringSet, lambda contract: to_string_set(contract.comboLegs)),
        ("DeltaNeutralContract", dtypes.string, lambda contract: to_string_val(contract.deltaNeutralContract)),
    ]


logger_contract = IbComplexTypeLogger("Contract", _details_contract())


####

def _details_contract_details() -> List[Tuple]:
    """Details for logging ContractDetails."""

    def map_null_int(value: int) -> Optional[int]:

        if value == 2147483647:
            return None

        return value

    def map_sec_id_list(value):
        if not value:
            return None

        return to_string_set([f"{v.tag}={v.value}" for v in value])

    return [
        *_include_details(_details_contract(), lambda cd: cd.contract),
        ("MarketName", dtypes.string, lambda cd: cd.marketName),
        ("MinTick", dtypes.float64, lambda cd: cd.minTick),
        ("OrderTypes", dtypes.StringSet, lambda cd: to_string_set(cd.orderTypes.split(","))),
        ("ValidExchanges", dtypes.StringSet, lambda cd: to_string_set(cd.validExchanges.split(","))),
        ("PriceMagnifier", dtypes.int64, lambda cd: cd.priceMagnifier),
        ("UnderConId", dtypes.int64, lambda cd: cd.underConId),
        ("LongName", dtypes.string, lambda cd: cd.longName),
        ("ContractMonth", dtypes.string, lambda cd: cd.contractMonth),
        ("Industry", dtypes.string, lambda cd: cd.industry),
        ("Category", dtypes.string, lambda cd: cd.category),
        ("SubCategory", dtypes.string, lambda cd: cd.subcategory),
        ("TimeZoneId", dtypes.string, lambda cd: cd.timeZoneId),
        ("TradingHours", dtypes.StringSet, lambda cd: to_string_set(cd.tradingHours.split(";"))),
        ("LiquidHours", dtypes.StringSet, lambda cd: to_string_set(cd.liquidHours.split(";"))),
        ("EvRule", dtypes.string, lambda cd: cd.evRule),
        ("EvMultiplier", dtypes.int64, lambda cd: cd.evMultiplier),
        ("AggGroup", dtypes.int64, lambda cd: map_null_int(cd.aggGroup)),
        ("UnderSymbol", dtypes.string, lambda cd: cd.underSymbol),
        ("UnderSecType", dtypes.string, lambda cd: cd.underSecType),
        ("MarketRuleIds", dtypes.StringSet, lambda cd: to_string_set(cd.marketRuleIds.split(","))),
        ("SecIdList", dtypes.StringSet, lambda cd: map_sec_id_list(cd.secIdList)),
        ("RealExpirationDate", dtypes.string, lambda cd: cd.realExpirationDate),
        ("LastTradeTime", dtypes.string, lambda cd: cd.lastTradeTime),
        ("StockType", dtypes.string, lambda cd: cd.stockType),
        # BOND values
        ("CUSIP", dtypes.string, lambda cd: cd.cusip),
        ("Ratings", dtypes.string, lambda cd: cd.ratings),
        ("DescAppend", dtypes.string, lambda cd: cd.descAppend),
        ("BondType", dtypes.string, lambda cd: cd.bondType),
        ("CouponType", dtypes.string, lambda cd: cd.couponType),
        ("Callable", dtypes.bool_, lambda cd: cd.callable),
        ("Putable", dtypes.bool_, lambda cd: cd.putable),
        ("Coupon", dtypes.double, lambda cd: float(cd.coupon)),
        ("Convertible", dtypes.bool_, lambda cd: cd.convertible),
        ("Maturity", dtypes.string, lambda cd: cd.maturity),
        # TODO: convert date time?  Values are not provided in TWS, and the format is not documented. (https://github.com/deephaven-examples/deephaven-ib/issues/10)
        ("IssueDate", dtypes.string, lambda cd: cd.issueDate),
        # TODO: convert date time?  Values are not provided in TWS, and the format is not documented. (https://github.com/deephaven-examples/deephaven-ib/issues/10)
        ("NextOptionDate", dtypes.string, lambda cd: cd.nextOptionDate),
        # TODO: convert date time?  Values are not provided in TWS, and the format is not documented. (https://github.com/deephaven-examples/deephaven-ib/issues/10)
        ("NextOptionType", dtypes.string, lambda cd: cd.nextOptionType),
        ("NextOptionPartial", dtypes.bool_, lambda cd: cd.nextOptionPartial),
        ("Notes", dtypes.string, lambda cd: cd.notes),
    ]


logger_contract_details = IbComplexTypeLogger("ContractDetails", _details_contract_details())


####

def _details_price_increment() -> List[Tuple]:
    """Details for logging PriceIncrement."""

    return [
        ("LowEdge", dtypes.float64, lambda pi: pi.lowEdge),
        ("Increment", dtypes.float64, lambda pi: pi.increment),
    ]


logger_price_increment = IbComplexTypeLogger("PriceIncrement", _details_price_increment())


####

def _details_bar_data() -> List[Tuple]:
    """ Details for logging BarData. """

    def map_null(val):
        if val <= 0:
            return None

        return val

    def parse_timestamp(bd):
        if len(bd.date) is 8:
            # bd.date is a date string
            year = bd.date[0:4]
            month = bd.date[4:6]
            day = bd.date[6:]
            time_string = f"{year}{month}{day} 23:59:59"
            return ib_to_j_instant(time_string)
        else:
            # bd.date is unix sec
            return unix_sec_to_j_instant(int(bd.date))

    return [
        ("Timestamp", dtypes.Instant, parse_timestamp),
        ("Open", dtypes.float64, lambda bd: bd.open),
        ("High", dtypes.float64, lambda bd: bd.high),
        ("Low", dtypes.float64, lambda bd: bd.low),
        ("Close", dtypes.float64, lambda bd: bd.close),
        ("Volume", dtypes.float64, lambda bd: map_null(bd.volume)),
        ("BarCount", dtypes.int64, lambda bd: map_null(bd.barCount)),
        ("WAP", dtypes.float64, lambda bd: map_null(bd.wap)),
    ]


logger_bar_data = IbComplexTypeLogger("BarData", _details_bar_data())


####

def _details_real_time_bar_data() -> List[Tuple]:
    """ Details for logging RealTimeBarData. """

    def map_null(val):
        if val <= 0:
            return None

        return val

    return [
        ("Timestamp", dtypes.Instant, lambda bd: unix_sec_to_j_instant(bd.time)),
        ("TimestampEnd", dtypes.Instant, lambda bd: unix_sec_to_j_instant(bd.endTime)),
        ("Open", dtypes.float64, lambda bd: bd.open_),
        ("High", dtypes.float64, lambda bd: bd.high),
        ("Low", dtypes.float64, lambda bd: bd.low),
        ("Close", dtypes.float64, lambda bd: bd.close),
        ("Volume", dtypes.float64, lambda bd: map_null(bd.volume)),
        ("WAP", dtypes.float64, lambda bd: map_null(bd.wap)),
        ("Count", dtypes.int64, lambda bd: map_null(bd.count)),
    ]


logger_real_time_bar_data = IbComplexTypeLogger("RealTimeBarData", _details_real_time_bar_data())


####

def _details_tick_attrib() -> List[Tuple]:
    """ Details for logging TickAttrib. """

    return [
        ("CanAutoExecute", dtypes.bool_, lambda ta: ta.canAutoExecute),
        ("PastLimit", dtypes.bool_, lambda ta: ta.pastLimit),
        ("PreOpen", dtypes.bool_, lambda ta: ta.preOpen),
    ]


logger_tick_attrib = IbComplexTypeLogger("TickAttrib", _details_tick_attrib())


###

def _details_tick_attrib_last() -> List[Tuple]:
    """Details for logging TickAttribLast."""

    return [
        ("PastLimit", dtypes.bool_, lambda ta: ta.pastLimit),
        ("Unreported", dtypes.bool_, lambda ta: ta.unreported),
    ]


logger_tick_attrib_last = IbComplexTypeLogger("TickAttribLast", _details_tick_attrib_last())


####

def _details_historical_tick_last() -> List[Tuple]:
    """Details for logging HistoricalTickLast."""

    # https://www.interactivebrokers.com/en/index.php?f=7235

    special_conditions_codes = {
        "B": "Average Price Trade",
        "Q": "Market Center Official Open",
        "C": "Cash Trade (Same Day Clearing)",
        "R": "Seller",
        "D": "Distribution",
        "T": "Extended Hours Trade",
        "E": "Automatic Execution",
        "U": "Extended Hours Sold (Out of Sequence)",
        "F": "Intermarket Sweep Order",
        "V": "Stock-Option Trade",
        "G": "Bunched Sold Trade",
        "X": "Cross Trade",
        "H": "Price Variation Trade",
        "Z": "Sold (Out of Sequence)",
        "I": "Odd Lot Trade",
        "4": "Derivatively priced",
        "K": "Rule 127 (NYSE only) or Rule 155 (NYSE MKT only)",
        "5": "Market Center Reopening Trade",
        "L": "Sold Last (Late Reporting)",
        "6": "Market Center Closing Trade",
        "M": "Market Center Official Close",
        "7": "Reserved",
        "N": "Next Day Trade (Next Day Clearing)",
        "8": "Reserved",
        "O": "Market Center Opening Trade",
        "9": "Corrected Consolidated Close Price as per Listing Market",
        "P": "Prior Reference Price",
    }

    def map_special_conditions(special_conditions: str) -> Any:
        if not special_conditions:
            return None

        return to_string_set([map_values(v, special_conditions_codes) for v in "".join(special_conditions.split())])


    return [
        ("Timestamp", dtypes.Instant, lambda t: unix_sec_to_j_instant(t.time)),
        ("Price", dtypes.float64, lambda t: t.price),
        ("Size", dtypes.float64, lambda t: t.size),
        *_include_details(_details_tick_attrib_last(), lambda t: t.tickAttribLast),
        ("Exchange", dtypes.string, lambda t: t.exchange),
        ("SpecialConditions", dtypes.StringSet, lambda t: map_special_conditions(t.specialConditions))
    ]


logger_hist_tick_last = IbComplexTypeLogger("HistoricalTickLast", _details_historical_tick_last())


####

def _details_tick_attrib_bid_ask() -> List[Tuple]:
    """Details for logging TickAttribBidAsk."""

    return [
        ("BidPastLow", dtypes.bool_, lambda ta: ta.bidPastLow),
        ("AskPastHigh", dtypes.bool_, lambda ta: ta.askPastHigh),
    ]


logger_tick_attrib_bid_ask = IbComplexTypeLogger("TickAttribBidAsk", _details_tick_attrib_bid_ask())


####

def _details_historical_tick_bid_ask() -> List[Tuple]:
    """Details for logging HistoricalTickBidAsk."""

    return [
        ("Timestamp", dtypes.Instant, lambda t: unix_sec_to_j_instant(t.time)),
        ("BidPrice", dtypes.float64, lambda t: t.priceBid),
        ("AskPrice", dtypes.float64, lambda t: t.priceAsk),
        ("BidSize", dtypes.float64, lambda t: t.sizeBid),
        ("AskSize", dtypes.float64, lambda t: t.sizeAsk),
        *_include_details(_details_tick_attrib_bid_ask(), lambda t: t.tickAttribBidAsk),
    ]


logger_hist_tick_bid_ask = IbComplexTypeLogger("HistoricalTickBidAsk", _details_historical_tick_bid_ask())


####

def _details_order() -> List[Tuple]:
    """ Details for logging Orders. """

    oca_types = {1: "CancelWithBlock", 2: "ReduceWithBlock", 3: "ReduceNonBlock"}
    trigger_methods = {0: "Default", 1: "DoubleBidAsk", 2: "Last", 3: "DoubleLast", 4: "BidAsk",
                       7: "LastOrBidAsk", 8: "MidPoint"}
    rule80_values = {"": None, "0": None, "I": "Individual", "A": "Agency", "W": "AgentOtherMember",
                     "J": "IndividualPTIA",
                     "U": "AgencyPTIA", "M": "AgentOtherMemberPTIA", "K": "IndividualPT", "Y": "AgencyPT",
                     "N": "AgentOtherMemberPT"}
    open_close_values = {"": None, "O": "Open", "C": "Close"}
    origin_values = {0: "Customer", 1: "Firm", 2: "Unknown"}
    short_sale_slot_values = {0: None, 1: "Holding", 2: "Elsewhere"}
    volatility_type = {0: None, 1: "Daily", 2: "Annual"}
    reference_price_type = {0: None, 1: "Average", 2: "BidOrAsk"}
    hedge_type = {"": None, "D": "Delta", "B": "Beta", "F": "FX", "P": "Pair"}
    auction_stragey_values = {0: "Unset", 1: "Match", 2: "Improvement", 3: "Transparent"}

    return [

        # order identifier
        ("OrderId", dtypes.int64, lambda o: o.orderId),
        ("ClientId", dtypes.int64, lambda o: o.clientId),
        ("PermId", dtypes.int64, lambda o: o.permId),

        # main order fields
        ("Action", dtypes.string, lambda o: o.action),
        ("TotalQuantity", dtypes.float64, lambda o: o.totalQuantity),
        ("OrderType", dtypes.string, lambda o: o.orderType),
        ("LmtPrice", dtypes.float64, lambda o: o.lmtPrice),
        ("AuxPrice", dtypes.float64, lambda o: o.auxPrice),

        # extended order fields
        ("TIF", dtypes.string, lambda o: o.tif),
        ("ActiveStartTime", dtypes.string, lambda o: o.activeStartTime),
        ("ActiveStopTime", dtypes.string, lambda o: o.activeStopTime),
        ("OcaGroup", dtypes.string, lambda o: o.ocaGroup),
        ("OcaType", dtypes.string, lambda o: map_values(o.ocaType, oca_types)),
        ("OrderRef", dtypes.string, lambda o: o.orderRef),
        ("Transmit", dtypes.bool_, lambda o: o.transmit),
        ("ParentId", dtypes.int64, lambda o: o.parentId),
        ("BlockOrder", dtypes.bool_, lambda o: o.blockOrder),
        ("SweepToFill", dtypes.bool_, lambda o: o.sweepToFill),
        ("DisplaySize", dtypes.int64, lambda o: o.displaySize),
        ("TriggerMethod", dtypes.string, lambda o: map_values(o.triggerMethod, trigger_methods)),
        ("OutsideRth", dtypes.bool_, lambda o: o.outsideRth),
        ("Hidden", dtypes.bool_, lambda o: o.hidden),
        ("GoodAfterTime", dtypes.string, lambda o: o.goodAfterTime),
        ("GoodTillDate", dtypes.string, lambda o: o.goodTillDate),
        ("Rule80A", dtypes.string, lambda o: map_values(o.rule80A, rule80_values)),
        ("AllOrNone", dtypes.bool_, lambda o: o.allOrNone),
        ("MinQty", dtypes.int64, lambda o: o.minQty),
        ("PercentOffset", dtypes.float64, lambda o: o.percentOffset),
        ("OverridePercentageConstraints", dtypes.bool_, lambda o: o.overridePercentageConstraints),
        ("TrailStopPrice", dtypes.float64, lambda o: o.trailStopPrice),
        ("TrailingPercent", dtypes.float64, lambda o: o.trailingPercent),

        # financial advisors only
        ("FaGroup", dtypes.string, lambda o: o.faGroup),
        ("FaProfile", dtypes.string, lambda o: o.faProfile),
        ("FaMethod", dtypes.string, lambda o: o.faMethod),
        ("FaPercentage", dtypes.string, lambda o: o.faPercentage),

        # institutional (ie non-cleared) only
        ("DesignatedLocation", dtypes.string, lambda o: o.designatedLocation),
        ("OpenClose", dtypes.string, lambda o: map_values(o.openClose, open_close_values)),
        ("Origin", dtypes.string, lambda o: map_values(o.origin, origin_values)),
        ("ShortSaleSlot", dtypes.string, lambda o: map_values(o.shortSaleSlot, short_sale_slot_values)),
        ("ExemptCode", dtypes.int64, lambda o: o.exemptCode),

        # SMART routing only
        ("DiscretionaryAmt", dtypes.float64, lambda o: o.discretionaryAmt),
        ("OptOutSmarRouting", dtypes.bool_, lambda o: o.optOutSmartRouting),

        # BOX exchange orders only
        ("AuctionStrategy", dtypes.string, lambda o: map_values(o.auctionStrategy, auction_stragey_values)),
        ("StartingPrice", dtypes.float64, lambda o: o.startingPrice),
        ("StockRefPrice", dtypes.float64, lambda o: o.stockRefPrice),
        ("Delta", dtypes.float64, lambda o: o.delta),

        # pegged to stock and VOL orders only
        ("StockRangeLower", dtypes.float64, lambda o: o.stockRangeLower),
        ("StockRangeUpper", dtypes.float64, lambda o: o.stockRangeUpper),

        ("RandomizePrice", dtypes.bool_, lambda o: o.randomizePrice),
        ("RandomizeSize", dtypes.bool_, lambda o: o.randomizeSize),

        # VOLATILITY ORDERS ONLY
        ("Volatility", dtypes.float64, lambda o: o.volatility),
        ("VolatilityType", dtypes.string, lambda o: map_values(o.volatilityType, volatility_type)),
        ("DeltaNeutralOrderType", dtypes.string, lambda o: o.deltaNeutralOrderType),
        ("DeltaNeutralAuxPrice", dtypes.float64, lambda o: o.deltaNeutralAuxPrice),
        ("DeltaNeutralConId", dtypes.int64, lambda o: o.deltaNeutralConId),
        ("DeltaNeutralSettlingFirm", dtypes.string, lambda o: o.deltaNeutralSettlingFirm),
        ("DeltaNeutralClearingAccount", dtypes.string, lambda o: o.deltaNeutralClearingAccount),
        ("DeltaNeutralClearingIntent", dtypes.string, lambda o: o.deltaNeutralClearingIntent),
        ("DeltaNeutralOpenClose", dtypes.string, lambda o: o.deltaNeutralOpenClose),
        ("DeltaNeutralShortSale", dtypes.bool_, lambda o: o.deltaNeutralShortSale),
        ("DeltaNeutralShortSaleSlot", dtypes.int64, lambda o: o.deltaNeutralShortSaleSlot),
        ("DeltaNeutralDesignatedLocation", dtypes.string, lambda o: o.deltaNeutralDesignatedLocation),
        ("ContinuousUpdate", dtypes.bool_, lambda o: o.continuousUpdate),
        ("ReferencePriceType", dtypes.string, lambda o: map_values(o.referencePriceType, reference_price_type)),

        # COMBO ORDERS ONLY
        ("BasisPoints", dtypes.float64, lambda o: o.basisPoints),
        ("BasisPointsType", dtypes.int64, lambda o: o.basisPointsType),

        # SCALE ORDERS ONLY
        ("ScaleInitLevelSize", dtypes.int64, lambda o: o.scaleInitLevelSize),
        ("ScaleSubsLevelSize", dtypes.int64, lambda o: o.scaleSubsLevelSize),
        ("ScalePriceIncrement", dtypes.float64, lambda o: o.scalePriceIncrement),
        ("ScalePriceAdjustValue", dtypes.float64, lambda o: o.scalePriceAdjustValue),
        ("ScalePriceAdjustInterval", dtypes.int64, lambda o: o.scalePriceAdjustInterval),
        ("ScaleProfitOffset", dtypes.float64, lambda o: o.scaleProfitOffset),
        ("ScaleAutoReset", dtypes.bool_, lambda o: o.scaleAutoReset),
        ("ScaleInitPosition", dtypes.int64, lambda o: o.scaleInitPosition),
        ("ScaleInitFillQty", dtypes.int64, lambda o: o.scaleInitFillQty),
        ("ScaleRandomPercent", dtypes.bool_, lambda o: o.scaleRandomPercent),
        ("ScaleTable", dtypes.string, lambda o: o.scaleTable),

        # HEDGE ORDERS
        ("HedgeType", dtypes.string, lambda o: map_values(o.hedgeType, hedge_type)),
        ("HedgeParam", dtypes.string, lambda o: o.hedgeParam),

        # Clearing info
        ("Account", dtypes.string, lambda o: o.account),
        ("SettlingFirm", dtypes.string, lambda o: o.settlingFirm),
        ("ClearingAccount", dtypes.string, lambda o: o.clearingAccount),
        ("ClearingIntent", dtypes.string, lambda o: o.clearingIntent),

        # ALGO ORDERS ONLY
        ("AlgoStrategy", dtypes.string, lambda o: o.algoStrategy),

        ("AlgoParams", dtypes.StringSet, lambda o: to_string_set(o.algoParams)),
        ("SmartComboRoutingParams", dtypes.StringSet, lambda o: to_string_set(o.smartComboRoutingParams)),

        ("AlgoId", dtypes.string, lambda o: o.algoId),

        # What-if
        ("WhatIf", dtypes.bool_, lambda o: o.whatIf),

        # Not Held
        ("NotHeld", dtypes.bool_, lambda o: o.notHeld),
        ("Solicited", dtypes.bool_, lambda o: o.solicited),

        # models
        ("ModelCode", dtypes.string, lambda o: o.modelCode),

        # order combo legs

        ("OrderComboLegs", dtypes.StringSet, lambda o: to_string_set(o.orderComboLegs)),

        ("OrderMiscOptions", dtypes.StringSet, lambda o: to_string_set(o.orderMiscOptions)),

        # VER PEG2BENCH fields:
        ("ReferenceContractId", dtypes.int64, lambda o: o.referenceContractId),
        ("PeggedChangeAmount", dtypes.float64, lambda o: o.peggedChangeAmount),
        ("IsPeggedChangeAmountDecrease", dtypes.bool_, lambda o: o.isPeggedChangeAmountDecrease),
        ("ReferenceChangeAmount", dtypes.float64, lambda o: o.referenceChangeAmount),
        ("ReferenceExchangeId", dtypes.string, lambda o: o.referenceExchangeId),
        ("AdjustedOrderType", dtypes.string, lambda o: o.adjustedOrderType),

        ("TriggerPrice", dtypes.float64, lambda o: o.triggerPrice),
        ("AdjustedStopPrice", dtypes.float64, lambda o: o.adjustedStopPrice),
        ("AdjustedStopLimitPrice", dtypes.float64, lambda o: o.adjustedStopLimitPrice),
        ("AdjustedTrailingAmount", dtypes.float64, lambda o: o.adjustedTrailingAmount),
        ("AdjustableTrailingUnit", dtypes.int64, lambda o: o.adjustableTrailingUnit),
        ("LmtPriceOffset", dtypes.float64, lambda o: o.lmtPriceOffset),

        ("Conditions", dtypes.StringSet, lambda o: to_string_set(o.conditions)),
        ("ConditionsCancelOrder", dtypes.bool_, lambda o: o.conditionsCancelOrder),
        ("ConditionsIgnoreRth", dtypes.bool_, lambda o: o.conditionsIgnoreRth),

        # ext operator
        ("ExtOperator", dtypes.string, lambda o: o.extOperator),

        # native cash quantity
        ("CashQty", dtypes.float64, lambda o: o.cashQty),

        ("Mifid2DecisionMaker", dtypes.string, lambda o: o.mifid2DecisionMaker),
        ("Mifid2DecisionAlgo", dtypes.string, lambda o: o.mifid2DecisionAlgo),
        ("Mifid2ExecutionTrader", dtypes.string, lambda o: o.mifid2ExecutionTrader),
        ("Mifid2ExecutionAlgo", dtypes.string, lambda o: o.mifid2ExecutionAlgo),

        ("DontUseAutoPriceForHedge", dtypes.bool_, lambda o: o.dontUseAutoPriceForHedge),

        ("IsOmsContainer", dtypes.bool_, lambda o: o.isOmsContainer),

        ("DiscretionaryUpToLimitPrice", dtypes.bool_, lambda o: o.discretionaryUpToLimitPrice),

        ("AutoCancelDate", dtypes.string, lambda o: o.autoCancelDate),
        ("FilledQuantity", dtypes.float64, lambda o: o.filledQuantity),
        ("RefFuturesConId", dtypes.int64, lambda o: o.refFuturesConId),
        ("AutoCancelParent", dtypes.bool_, lambda o: o.autoCancelParent),
        ("Shareholder", dtypes.string, lambda o: o.shareholder),
        ("ImbalanceOnly", dtypes.bool_, lambda o: o.imbalanceOnly),
        ("RouteMarketableToBbo", dtypes.bool_, lambda o: o.routeMarketableToBbo),
        ("ParentPermId", dtypes.int64, lambda o: o.parentPermId),

        ("UsePriceMgmtAlgo", dtypes.bool_, lambda o: o.usePriceMgmtAlgo),

        # soft dollars
        ("SoftDollarTier", dtypes.string, lambda o: to_string_val(o.softDollarTier)),
    ]


logger_order = IbComplexTypeLogger("Order", _details_order())


####

def _details_order_state() -> List[Tuple]:
    """ Details for logging OrderState. """

    return [
        ("Status", dtypes.string, lambda os: os.status),

        ("InitMarginBefore", dtypes.string, lambda os: os.initMarginBefore),
        ("MaintMarginBefore", dtypes.string, lambda os: os.maintMarginBefore),
        ("EquityWithLoanBefore", dtypes.string, lambda os: os.equityWithLoanBefore),
        ("InitMarginChange", dtypes.string, lambda os: os.initMarginChange),
        ("MaintMarginChange", dtypes.string, lambda os: os.maintMarginChange),
        ("EquityWithLoanChange", dtypes.string, lambda os: os.equityWithLoanChange),
        ("InitMarginAfter", dtypes.string, lambda os: os.initMarginAfter),
        ("MaintMarginAfter", dtypes.string, lambda os: os.maintMarginAfter),
        ("EquityWithLoanAfter", dtypes.string, lambda os: os.equityWithLoanAfter),

        ("Commission", dtypes.float64, lambda os: os.commission),
        ("MinCommission", dtypes.float64, lambda os: os.minCommission),
        ("MaxCommission", dtypes.float64, lambda os: os.maxCommission),
        ("CommissionCurrency", dtypes.string, lambda os: os.commissionCurrency),
        ("WarningText", dtypes.string, lambda os: os.warningText),
        ("CompletedTime", dtypes.string, lambda os: os.completedTime),
        ("CompletedStatus", dtypes.string, lambda os: os.completedStatus),
    ]


logger_order_state = IbComplexTypeLogger("OrderState", _details_order_state())


####

def _details_execution() -> List[Tuple]:
    """ Details for logging Execution. """

    return [
        ("ExecId", dtypes.string, lambda e: e.execId),
        ("Timestamp", dtypes.Instant, lambda e: ib_to_j_instant(e.time)),
        ("AcctNumber", dtypes.string, lambda e: e.acctNumber),
        ("Exchange", dtypes.string, lambda e: e.exchange),
        ("Side", dtypes.string, lambda e: e.side),
        ("Shares", dtypes.float64, lambda e: e.shares),
        ("Price", dtypes.float64, lambda e: e.price),
        ("PermId", dtypes.int64, lambda e: e.permId),
        ("ClientId", dtypes.int64, lambda e: e.clientId),
        ("OrderId", dtypes.int64, lambda e: e.orderId),
        ("Liquidation", dtypes.int64, lambda e: e.liquidation),
        ("CumQty", dtypes.float64, lambda e: e.cumQty),
        ("AvgPrice", dtypes.float64, lambda e: e.avgPrice),
        ("OrderRef", dtypes.string, lambda e: e.orderRef),
        ("EvRule", dtypes.string, lambda e: e.evRule),
        ("EvMultiplier", dtypes.float64, lambda e: e.evMultiplier),
        ("ModelCode", dtypes.string, lambda e: e.modelCode),
        ("LastLiquidity", dtypes.int64, lambda e: e.lastLiquidity),
    ]


logger_execution = IbComplexTypeLogger("Execution", _details_execution())

####

def _details_commission_report() -> List[Tuple]:
    """ Details for logging CommissionReport. """

    def format_yield_redemption_date(date: int) -> Optional[str]:
        if date == 0:
            return None

        # YYYYMMDD format
        d = date % 100
        m = int((date / 100) % 100)
        y = int(date / 10000)
        return f"{y:04}-{m:02}-{d:02}"

    def map_null_value(value: float) -> Optional[float]:

        if value == sys.float_info.max:
            return None

        return value


    return [
        ("ExecId", dtypes.string, lambda cr: cr.execId),
        ("Commission", dtypes.float64, lambda cr: cr.commission),
        ("Currency", dtypes.string, lambda cr: cr.currency),
        ("RealizedPNL", dtypes.float64, lambda cr: map_null_value(cr.realizedPNL)),
        ("Yield", dtypes.float64, lambda cr: map_null_value(cr.yield_)),
        ("YieldRedemptionDate", dtypes.string, lambda cr: format_yield_redemption_date(cr.yieldRedemptionDate)),
    ]


logger_commission_report = IbComplexTypeLogger("CommissionReport", _details_commission_report())


####

def _details_news_provider() -> List[Tuple]:
    """ Details for logging NewsProvider. """

    return [
        ("Code", dtypes.string, lambda np: np.code),
        ("Name", dtypes.string, lambda np: np.name),
    ]


logger_news_provider = IbComplexTypeLogger("NewsProvider", _details_news_provider())

####



"""Functionality for working with stack traces."""

import traceback
import threading
import sys

def trace_str() -> str:
    """Gets a string of the current stacktrace."""
    return "".join(traceback.format_stack())


def trace_thread_str(thread:threading.Thread) -> str:
    """Gets a string of the stacktrace of a thread."""
    try:
        return "".join(traceback.format_stack(sys._current_frames()[thread.ident]))
    except KeyError:
        return f"Thread stack not found: thread={thread.ident}"


def trace_all_threads_str() -> str:
    """Get the stacktraces for all threads as a string."""

    rst = "Stack Traces:\n"

    for th in threading.enumerate():
        rst += str(th)
        rst += trace_thread_str(th)
        rst += "\n"

    return rst


import os
import re

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

dh_ib_version = os.getenv("DH_IB_VERSION")

if not dh_ib_version:
    raise Exception("deephaven-ib version must be set via the DH_IB_VERSION environment variable.")

ib_version = os.getenv("IB_VERSION")

if not ib_version:
    raise Exception("ibapi version must be set via the IB_VERSION environment variable.")


def is_valid_semver(version: str, allow_zero_prefix: bool=False):
    """
    Checks if a string is in valid semver format.

    Args:
        version: The version string to validate.
        allow_zero_prefix: Allow zero prefixes

    Returns:
        True if the string is in valid semver format, False otherwise.
    """
    if allow_zero_prefix:
        pattern = r'^([0-9]\d*)\.([0-9]\d*)\.([0-9]\d*)(-([0-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(\.([0-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*)?(\+[0-9a-zA-Z-]+(\.[0-9a-zA-Z-]+)*)?$'
    else:
        pattern = r'^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(-(0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(\.(0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*)?(\+[0-9a-zA-Z-]+(\.[0-9a-zA-Z-]+)*)?$'

    return bool(re.match(pattern, version))


def version_assert_format(version: str, allow_zero_prefix: bool=False) -> None:
    """Assert that a version string is formatted correctly.

    Args:
        version: The version string to check.
        allow_zero_prefix: Allow zero prefixes

    Raises:
        ValueError: If the version string is not formatted correctly.
    """
    if not version:
        raise ValueError("Version string is empty.")

    if not is_valid_semver(version, allow_zero_prefix=allow_zero_prefix):
        raise ValueError(f"Version string is not in semver format: {version}")


version_assert_format(dh_ib_version)
version_assert_format(ib_version, allow_zero_prefix=True)

setuptools.setup(
    name="deephaven_ib",
    version=dh_ib_version,
    author="David R. (Chip) Kent IV",
    author_email="chipkent@deephaven.io",
    description="An Interactive Brokers integration for Deephaven",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/deephaven-examples/deephaven-ib",
    project_urls={
        "Deephaven": "https://deephaven.io",
        "Deephaven GitHub": "https://github.com/deephaven/deephaven-core",
        "GitHub Issues": "https://github.com/deephaven-examples/deephaven-ib/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Financial and Insurance Industry",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Office/Business :: Financial :: Investment",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.10",
    install_requires=[
        "deephaven-server",
        "pandas",
        f"ibapi=={ib_version}",
        "lxml",
        "ratelimit",
    ],
)

